#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "button.h"
#include "edent.h"
#include "edentity.h"
#include "entity.h"
#include "error.h"
#include "mouse.h"
#include "quest.h"
#include "qui.h"
#include "video.h"


/* TODO: get rid of MAX_MODELS, clean this up */
#define MAX_MODELS 512

static QUI_window_t *w;

static int base_ent;
static int active_ent;
static int models;

static entity_t *ents[MAX_MODELS];
static int brushes[MAX_MODELS];


static void ClearModelList(void)
{
   int i;
   /* Clear text area */
   for (i=w->pos.y+51;i<w->pos.y+293;i++)
      DrawLine(w->pos.x+11,i,w->pos.x+359,i,BG_COLOR);
}

static void DrawModelList(void)
{
   int i,j;
	/* Draw model list */
	for (i=base_ent,j=0;(j<17) && (i<models);i++,j++)
   {
		QUI_DrawStr(w->pos.x+15,w->pos.y+52+(j*14),
		            BG_COLOR,(i==active_ent)?14:0,
		            0,0,"%s",GetKeyValue(ents[i],"classname"));
		QUI_DrawStr(w->pos.x+300,w->pos.y+52+(j*14),
		            BG_COLOR,(i==active_ent)?14:0,
		            0,0,"%i",brushes[i]);
   }
}

static void ReadModelList(void)
{
   brush_t *b;
   entity_t *e;

   models=0;
   for (e=M.EntityHead;e;e=e->Next)
   {
      if (!GetKeyValue(e,"origin"))
      {
         ents[models]=e;
         brushes[models]=0;
         for (b=M.BrushHead;b;b=b->Next)
         {
            if (b->EntityRef==e)
               brushes[models]++;
         }
         models++;
         if (models==MAX_MODELS)
         {
            HandleError("model_list","Too many models");
            return;
         }
      }
   }
}


static void C_model_list(void)
{
   entity_t *e;

	unsigned char *temp_buf;

	int i,j;

   int b_ok,b_edit,b_delete;
   int b_tiny_up,b_tiny_down;
   int bp;


   active_ent=-1;

   ReadModelList();

	/* Set up window Position */
	w=&Q.window[POP_WINDOW_1+Q.num_popups];

	w->size.x=400;
	w->size.y=350;
	w->pos.x=(video.ScreenWidth-w->size.x)/2;
	w->pos.y=(video.ScreenHeight-w->size.y)/2;

	/* Set up some buttons */
   PushButtons();
	b_ok    =AddButtonText(0,0,B_ENTER|B_ESCAPE,"OK");
	b_edit  =AddButtonText(0,0,0,"Edit");
	b_delete=AddButtonText(0,0,0,"Delete");

	MoveButton(b_ok,
	           w->pos.x+50,
              w->pos.y+w->size.y-30);
	MoveButton(b_edit,
	           button[b_ok].x+button[b_ok].sx+5,
	           w->pos.y+w->size.y-30);
	MoveButton(b_delete,
	           button[b_edit].x+button[b_edit].sx+5,
              w->pos.y+w->size.y-30);

	base_ent=0;

	/* Set up scrolling buttons */
	b_tiny_up  =AddButtonPic(0,0,B_RAPID,"button_tiny_up");
	b_tiny_down=AddButtonPic(0,0,B_RAPID,"button_tiny_down");
	MoveButton(b_tiny_up,
              w->pos.x+w->size.x-30,
              w->pos.y+52);
	MoveButton(b_tiny_down,
	           w->pos.x+w->size.x-30,
	           w->pos.y+w->size.y-70);

	/* Actually draw the window */
	QUI_PopUpWindow(POP_WINDOW_1+Q.num_popups,"Models",&temp_buf);
	Q.num_popups++;

	/* Box Model selection area & Titles */
	QUI_DrawStr(w->pos.x+15,w->pos.y+33,
	            BG_COLOR,14,0,0,
	            "classname");
	QUI_DrawStr(w->pos.x+300,w->pos.y+33,
	            BG_COLOR,14,0,0,
	            "Brushes");

   DrawLine(w->pos.x+10,w->pos.y+50,
	         w->pos.x+10,w->pos.y+294,
	         4);
	DrawLine(w->pos.x+10,w->pos.y+50,
	         w->pos.x+360,w->pos.y+50,
	         4);
	DrawLine(w->pos.x+10,w->pos.y+294,
	         w->pos.x+360,w->pos.y+294,
	         8);
	DrawLine(w->pos.x+360,
	         w->pos.y+50,w->pos.x+360,
	         w->pos.y+294,
	         8);

	/* Draw Buttons */
   DrawButtons();

   DrawModelList();

	RefreshPart(w->pos.x, w->pos.y, w->pos.x+w->size.x, w->pos.y+w->size.y);

	while (1)
	{
      UpdateMouse();
      bp=UpdateButtons();

		if (bp!=-1)
		{
         if (bp==b_ok)
         {
				break;
			}

			if (bp==b_edit)
         {
				if (active_ent==-1)
            {
					HandleError("model_list","No model selected");
            }
				else
				{
               EditAnyEntity(1,&ents[active_ent]);

               ClearModelList();
               DrawModelList();

               RefreshPart(w->pos.x, w->pos.y, w->pos.x+w->size.x, w->pos.y+w->size.y);
               UpdateMouse();
               DrawMouse(mouse.x, mouse.y);
				}
			}

         if (bp==b_delete)
         {
				if (active_ent == -1)
            {
					HandleError("model_list","No model selected");
            }
				else
				{
               e=ents[active_ent];

               /* Make sure the entity we're about to delete isn't selected,
                should probably be done in DeleteEntity() */
               ClearSelEnts();

               DeleteEntity(e);

               ReadModelList();
               ClearModelList();
               DrawModelList();

               RefreshPart(w->pos.x, w->pos.y, w->pos.x+w->size.x, w->pos.y+w->size.y);
               UpdateMouse();
               DrawMouse(mouse.x, mouse.y);
				}
			}

         if (bp==b_tiny_up)
         {
            if (base_ent > 0)
               base_ent--;

            ClearModelList();
            DrawModelList();

            RefreshPart(w->pos.x, w->pos.y,
                        w->pos.x+w->size.x, w->pos.y+w->size.y);
            UpdateMouse();
            DrawMouse(mouse.x, mouse.y);
			}

			if (bp==b_tiny_down)
         {
            if ((base_ent+17) < models)
               base_ent++;

            ClearModelList();
            DrawModelList();

            RefreshPart(w->pos.x, w->pos.y,
                        w->pos.x+w->size.x, w->pos.y+w->size.y);
            UpdateMouse();
            DrawMouse(mouse.x, mouse.y);
         }
      }
      else // No button pressed (b==-1)
      if (mouse.button&1)
      {
			/* Check for click in model area */
			if (InBox(w->pos.x + 10,  w->pos.y + 50,
			          w->pos.x + 360, w->pos.y + 294))
			{
				/* Find which model they clicked on */
				for (i=base_ent, j=0; (j<17) && (i<models); i++, j++)
				{
					if (InBox(w->pos.x + 10,  w->pos.y + 52 + j*14,
					          w->pos.x + 360, w->pos.y + 66 + j*14))
						active_ent = i;
				}

            DrawModelList();

				RefreshPart(w->pos.x, w->pos.y,
				            w->pos.x+w->size.x, w->pos.y+w->size.y);
			}
		}
	}

	/* Pop down the window */
	Q.num_popups--;
	QUI_PopDownWindow(POP_WINDOW_1 + Q.num_popups, &temp_buf);

	RemoveButton(b_ok);
	RemoveButton(b_edit);
	RemoveButton(b_delete);
	RemoveButton(b_tiny_up);
	RemoveButton(b_tiny_down);
   PopButtons();

	RefreshPart(w->pos.x, w->pos.y, w->pos.x+w->size.x, w->pos.y+w->size.y);

	Cmd_SetRedraw(2);
#undef MAX_MODELS
}


static cmd_reg_t entlist_cmds[]={
{"model_list",C_model_list,"Show a list of all models in the map."},
};

static void RegCmds(void) __attribute__ ((constructor));
static void RegCmds(void)
{
   Cmd_Register(entlist_cmds,sizeof(entlist_cmds));
}

