#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "3d.h"
#include "bsp.h"
#include "color.h"
#include "dvport.h"
#include "menu.h"
#include "message.h"
#include "mouse.h"
#include "quest.h"
#include "qui.h"
#include "status.h"
#include "video.h"


static void C_toggle_fullbright(void)
{
   if (!M.display.vport[M.display.active_vport].fullbright)
      M.display.vport[M.display.active_vport].fullbright = TRUE;
   else
      M.display.vport[M.display.active_vport].fullbright = FALSE;

   NewMessage("Fullbright mode in viewport %d turned %s.",
      M.display.active_vport,
      M.display.vport[M.display.active_vport].fullbright?"on":"off");

   Cmd_SetRedraw(1);
}

static void C_toggle_axisaligned(void)
{
   if (M.display.vport[M.display.active_vport].mode==NOPERSP)
      return;

   if (M.display.vport[M.display.active_vport].axis_aligned)
      M.display.vport[M.display.active_vport].axis_aligned=0;
   else
      M.display.vport[M.display.active_vport].axis_aligned=1;
   NewMessage("Axis aligned mode in viewport %i turned %s.",
      M.display.active_vport,
      M.display.vport[M.display.active_vport].axis_aligned?"on":"off");
   Cmd_SetRedraw(1);
}


static void C_toggle_full_viewport(void)
{
   if (M.display.vp_full)
      M.display.vp_full=0;
   else
      M.display.vp_full=1;
   UpdateViewportPositions();
   Cmd_SetRedraw(2);
}

/* TODO: generalize */
static void C_num_vports(void)
{
   if (M.display.num_vports==4)
      InitViewports();
   else
   {
      M.display.num_vports=4;
      M.display.vport[0].f_xmin=0;
      M.display.vport[0].f_ymin=0;
      M.display.vport[0].f_xmax=0.5;
      M.display.vport[0].f_ymax=0.5;

      M.display.vport[1].f_xmin=0;
      M.display.vport[1].f_ymin=0.5;
      M.display.vport[1].f_xmax=0.5;
      M.display.vport[1].f_ymax=1;

      M.display.vport[2].f_xmin=0.5;
      M.display.vport[2].f_ymin=0;
      M.display.vport[2].f_xmax=1;
      M.display.vport[2].f_ymax=0.5;

      M.display.vport[3].f_xmin=0.5;
      M.display.vport[3].f_ymin=0.5;
      M.display.vport[3].f_xmax=1;
      M.display.vport[3].f_ymax=1;
      UpdateViewportPositions();
   }
   Cmd_SetRedraw(2);
}


static void C_depthclip_inc(void)
{
   status.depth_clip *= 1.2;
   Cmd_SetRedraw(2);
}

static void C_depthclip_dec(void)
{
   status.depth_clip /= 1.2;
   if (status.depth_clip < 10) status.depth_clip = 10;
   Cmd_SetRedraw(2);
}


static void C_toggle_2d(void)
{
   if (M.display.vport[M.display.active_vport].mode != NOPERSP)
   {
      M.display.vport[M.display.active_vport].mode = NOPERSP;
      M.display.vport[M.display.active_vport].axis_aligned=1;
   }
   else
   {
      M.display.vport[M.display.active_vport].mode = WIREFRAME;
   }
   Cmd_SetRedraw(1);
}

static void C_toggle_poly(void)
{
   if (M.display.vport[M.display.active_vport].mode != SOLID)
   {
      SetPal(PAL_TEXTURE);
      M.display.vport[M.display.active_vport].mode = SOLID;
   }
   else
   {
      SetPal(PAL_QUEST);
      M.display.vport[M.display.active_vport].mode = WIREFRAME;
   }
   Cmd_SetRedraw(1);
}

static void C_toggle_bsp(void)
{
   if ((M.display.vport[M.display.active_vport].mode != BSPVIEW) ||
        M.display.vport[M.display.active_vport].textured!=BSP_COL)
   {
      M.display.vport[M.display.active_vport].mode = BSPVIEW;
      M.display.vport[M.display.active_vport].textured=BSP_COL;
   }
   else
   {
      M.display.vport[M.display.active_vport].mode = WIREFRAME;
   }
   SetPal(PAL_QUEST);
   Cmd_SetRedraw(1);
}

static void C_toggle_flat_shaded(void)
{
   if ((M.display.vport[M.display.active_vport].textured==BSP_GREY) &&
       (M.display.vport[M.display.active_vport].mode==BSPVIEW))
   {
      M.display.vport[M.display.active_vport].textured=0;
      M.display.vport[M.display.active_vport].mode=WIREFRAME;
   }
   else
   {
      M.display.vport[M.display.active_vport].mode = BSPVIEW;
      M.display.vport[M.display.active_vport].textured=BSP_GREY;
   }
   SetPal(PAL_QUEST);
   Cmd_SetRedraw(1);
}

#define C_t(x,y) \
   static void C_toggle_##x(void) \
   { \
      if ((M.display.vport[M.display.active_vport].textured==BSP_TEX) && \
          (M.display.vport[M.display.active_vport].mode==y)) \
      { \
         M.display.vport[M.display.active_vport].textured=0; \
         M.display.vport[M.display.active_vport].mode=WIREFRAME; \
         SetPal(PAL_QUEST); \
      } \
      else \
      { \
         M.display.vport[M.display.active_vport].mode = BSPVIEW; \
         M.display.vport[M.display.active_vport].textured=y; \
         SetPal(PAL_TEXTURE); \
      } \
      Cmd_SetRedraw(1); \
   }

C_t(textured,BSP_TEX)
C_t(textured_correct,BSP_TEXC)
C_t(textured_light,BSP_LIGHT)
C_t(textured_shadow,BSP_LIGHTS)


static void C_bsp_rebuild(void)
{
   RebuildBSP();
   M.display.vport[M.display.active_vport].mode = BSPVIEW;
   Cmd_SetRedraw(2);
}


static void C_pal_game(void)
{
   SetPal(PAL_TEXTURE);
   Cmd_SetRedraw(2);
}

static void C_pal_quest(void)
{
   SetPal(PAL_QUEST);
   Cmd_SetRedraw(2);
}


static void C_vid_reinit(void)
{
   SetMode(options.vid_mode);
   memset(video.ScreenBuffer,0,video.ScreenBufferSize);
   InitMouse();
   SetPal(PAL_QUEST);
   QUI_RedrawWindow(MENU_WINDOW);
   QUI_RedrawWindow(MESG_WINDOW);
   QUI_RedrawWindow(TOOL_WINDOW);
   QUI_RedrawWindow(STATUS_WINDOW);
   UpdateAllViewports();
   if (!status.pop_menu)
      PopUpMenuWin();
   RefreshScreen();
}


/* TODO: move pts/lin handling to own file? along with BSP leak checker? */
static void C_toggle_pts(void)
{
   if (status.draw_pts)
      status.draw_pts=0;
   else
      status.draw_pts=1;

   NewMessage("Pts/lin drawing turned %s",
      status.draw_pts?"on":"off");

   Cmd_SetRedraw(2);
}


static cmd_reg_t vid_cmds[]={
{"vid_toggle_fullbright",C_toggle_fullbright,"Toggle fullbright."},
{"vid_toggle_axisaligned",C_toggle_axisaligned,"Toggle axis-aligned mode."},
{"vid_toggle_full_viewport",C_toggle_full_viewport,"Toggle full viewport."},
{"vid_num_vports",C_num_vports,"Switch number of viewports 3/4."},
{"vid_depthclip_inc",C_depthclip_inc,"Increase depth clipping distance."},
{"vid_depthclip_dec",C_depthclip_dec,"Decrease depth clipping distance."},

{"vid_toggle_2d"              ,C_toggle_2d              ,"Toggle flat 2d mode."},
{"vid_toggle_poly"            ,C_toggle_poly            ,"Toggle solid textured polygon mode."},
{"vid_toggle_bsp"             ,C_toggle_bsp             ,"Toggle BSP mode."},
{"vid_toggle_flat_shaded"     ,C_toggle_flat_shaded     ,"Toggle grey-shaded BSP mode."},
{"vid_toggle_textured"        ,C_toggle_textured        ,"Toggle textured BSP mode."},
{"vid_toggle_textured_correct",C_toggle_textured_correct,"Toggle perspective correct textured BSP mode."},
{"vid_toggle_textured_light"  ,C_toggle_textured_light  ,"Toggle lighted BSP mode."},
{"vid_toggle_textured_shadow" ,C_toggle_textured_shadow ,"Toggle lighted BSP mode with shadows."},

{"vid_bsp_rebuild",C_bsp_rebuild,"Rebuild the BSP tree."},
{"vid_pal_game",C_pal_game,"Switch to the current game's palette."},
{"vid_pal_quest",C_pal_quest,"Switch to Quest's palette."},
{"vid_reinit",C_vid_reinit,"Reinitialize the video system."},

{"toggle_pts",C_toggle_pts,"Toggle display of .pts/.lin data."},
};

static void RegCmds(void) __attribute__ ((constructor));
static void RegCmds(void)
{
   Cmd_Register(vid_cmds,sizeof(vid_cmds));
}

