#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "config.h"
#include "brush.h"
#include "camera.h"
#include "edbrush.h"
#include "geom.h"
#include "quest.h"
#include "s_3d.h"
#include "s_message.h"
#include "status.h"


/* Camera commands */

static void C_rot_up(void)
{
   Cam_LookUp();
   Cmd_SetRedraw(1);
}

static void C_rot_down(void)
{
   Cam_LookDown();
   Cmd_SetRedraw(1);
}

static void C_rot_left(void)
{
   Cam_TurnLeft();
   Cmd_SetRedraw(1);
}

static void C_rot_right(void)
{
   Cam_TurnRight();
   Cmd_SetRedraw(1);
}

static void C_rot_rleft(void)
{
   Cam_RollLeft();
   Cmd_SetRedraw(1);
}

static void C_rot_rright(void)
{
   Cam_RollRight();
   Cmd_SetRedraw(1);
}

static void C_move_forward(void)
{
   MoveCamera(S_cur_vport,MOVE_FORWARD);
   Cmd_SetRedraw(1);
}

static void C_move_backward(void)
{
   MoveCamera(S_cur_vport,MOVE_BACKWARD);
   Cmd_SetRedraw(1);
}

static void C_move_left(void)
{
   MoveCamera(S_cur_vport,MOVE_LEFT);
   Cmd_SetRedraw(1);
}

static void C_move_right(void)
{
   MoveCamera(S_cur_vport,MOVE_RIGHT);
   Cmd_SetRedraw(1);
}

static void C_move_down(void)
{
   MoveCamera(S_cur_vport,MOVE_DOWN);
   Cmd_SetRedraw(1);
}

static void C_move_up(void)
{
   MoveCamera(S_cur_vport,MOVE_UP);
   Cmd_SetRedraw(1);
}

static void C_zoom_in(void)
{
   S_cur_vport->zoom_amt *= cv_cam_zoom_speed->f;

   if (S_cur_vport->zoom_amt>12)
      S_cur_vport->zoom_amt=12;
   if (S_cur_vport->zoom_amt<0.125/20)
      S_cur_vport->zoom_amt=0.125/20;

   Cmd_SetRedraw(1);
}

static void C_zoom_out(void)
{
   S_cur_vport->zoom_amt /= cv_cam_zoom_speed->f;

   if (S_cur_vport->zoom_amt>12)
      S_cur_vport->zoom_amt=12;
   if (S_cur_vport->zoom_amt<0.125/20)
      S_cur_vport->zoom_amt=0.125/20;

   Cmd_SetRedraw(1);
}


static void C_cam_reinit(void)
{
   InitViewports();
   S_NewMessage("Camera Re-Initialized");
   Cmd_SetRedraw(2);
}


static void C_zoom_to(void)
{
	bsel_t *bs;
   esel_t *es;
   fsel_t *fs;

   brush_t *b;
   face_t *f;
   
   vec3_t min,max;
   int found;

   int i;

   found=0;

	for (bs=M.sel.bsel;bs;bs=bs->Next)
	{
      b=bs->Brush;
		for (i=0;i<b->num_verts;i++)
		{
         if (!found)
         {
            found=1;
            max=min=b->verts[i];
         }
         else
         {
            VEC_CheckMin(min,b->verts[i]);
            VEC_CheckMax(max,b->verts[i]);
         }
		}
	}
	for (fs=M.sel.fsel;fs;fs=fs->Next)
	{
      b=fs->Brush;
      f=&b->faces[fs->num];
		for (i=0;i<f->num_verts;i++)
		{
         if (!found)
         {
            found=1;
            max=min=b->verts[f->verts[i]];
         }
         else
         {
            VEC_CheckMin(min,b->verts[f->verts[i]]);
            VEC_CheckMax(max,b->verts[f->verts[i]]);
         }
		}
	}
   for (es=M.sel.esel;es;es=es->Next)
   {
      if (!found)
      {
         found=1;
         max=min=es->Entity->center;
      }
      else
      {
         VEC_CheckMin(min,es->Entity->center);
         VEC_CheckMax(max,es->Entity->center);
      }
   }

   if (!found)
   {
      S_NewMessage("cam.zoom_to: Nothing selected!");
      return;
   }

   min.x=(min.x+max.x)/2;
   min.y=(min.y+max.y)/2;
   min.z=(min.z+max.z)/2;
   AutoZoom(min);
   Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"cam.turn_right",C_rot_right ,{NULL,"alt+right"},{CMD_MOVE},"Rotate camera right."},
{"cam.turn_left" ,C_rot_left  ,{NULL,"alt+left" },{CMD_MOVE},"Rotate camera left." },
{"cam.look_up"   ,C_rot_up    ,{NULL,"alt+up"   },{CMD_MOVE},"Rotate camera up."   },
{"cam.look_down" ,C_rot_down  ,{NULL,"alt+down" },{CMD_MOVE},"Rotate camera down." },
{"cam.roll_right",C_rot_rright,{NULL,"d"        },{CMD_MOVE},"Roll camera right."  },
{"cam.roll_left" ,C_rot_rleft ,{NULL,"c"        },{CMD_MOVE},"Roll camera left."   },

{"cam.move_forward" ,C_move_forward ,{NULL,"a"    },{CMD_MOVE},"Move camera forwards." },
{"cam.move_backward",C_move_backward,{NULL,"z"    },{CMD_MOVE},"Move camera backwards."},
{"cam.move_left"    ,C_move_left    ,{NULL,"left" },{CMD_MOVE},"Move camera left."     },
{"cam.move_right"   ,C_move_right   ,{NULL,"right"},{CMD_MOVE},"Move camera right."    },
{"cam.move_up"      ,C_move_up      ,{NULL,"up"   },{CMD_MOVE},"Move camera up."       },
{"cam.move_down"    ,C_move_down    ,{NULL,"down" },{CMD_MOVE},"Move camera down."     },

{"cam.zoom_in" ,C_zoom_in ,{NULL,"s"},{CMD_MOVE},"Zoom camera in." },
{"cam.zoom_out",C_zoom_out,{NULL,"x"},{CMD_MOVE},"Zoom camera out."},

{"cam.zoom_to" ,C_zoom_to,{"Camera/Zoom to","f2"},{},"Centers all viewports on the current selection."},

{"cam.reinit",C_cam_reinit,{"Camera/Reset","control+alt+c"},{},"Reset camera."},
};

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/camera","Fundamental camera commands (moving around, rotating, zooming)."}};

PLUGIN_FOOT_CMDS

