#include <stdio.h>
#include <stdlib.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "csg.h"
#include "edsel.h"
#include "group.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"


static void AddBrushes(brush_t *b1, brush_t *b2)
{
	int i,j;
	cbrush_t A;
	brush_t *b;

   if (b1->bt->type!=BRUSH_NORMAL || b2->bt->type!=BRUSH_NORMAL)
   {
      S_Error("CSG Intersect","Can only use normal brushes for CSG operations!");
      return;
   }

	A.num_faces = b1->num_faces + b2->num_faces;
	A.faces = (cface_t *)Q_malloc(sizeof(cface_t) * A.num_faces);

   if (!A.faces)
   {
      S_Error("CSG Intersect","Out of memory!");
      return;
   }

	for (i=j=0;i<b1->num_faces;i++,j++)
      if (!CSG_FaceToCFace(&A.faces[j],b1,i))
      {
         S_Error("CSG Intersect","Out of memory!");
         return;
      }

	for (i=0;i<b2->num_faces;i++,j++)
      if (!CSG_FaceToCFace(&A.faces[j],b2,i))
      {
         S_Error("CSG Intersect","Out of memory!");
         return;
      }

   b=CSG_BuildBrush(NULL,&A);
   if (!b)
   {
      S_Error("CSG Intersect","Can't create new brush! (brushes don't intersect?)");
      return;
   }

	ClearSelVerts();
   ClearSelBrushes();

   b->Entity=M.WorldSpawn;
   b->Group=FindVisGroup(b1->Group);
   B_Link(b);
   AddSelBrush(b);
}


static void C_intersect(void)
{
   if (M.sel.num_bselected!=2)
   {
  	   S_Error("CSG Intersect","Must have exactly 2 brushes selected!");
      return;
   }

   AddBrushes(M.sel.bsel->Brush,M.sel.bsel->Next->Brush);

   Cmd_SetRedraw(2);
}


static const char help[]=
"This does a CSG intersection of two selected brushes (which must "
"intersect), i.e. it creates a new brush that covers the space where "
"the two brushes intersect.
";

static cmd_reg_t cmds[]={
{"csg.intersect",C_intersect,{"Brush/CSG/Intersect","o"},{},"CSG intersect",help},
};

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/csg/intersect","CSG intersect command."}};

PLUGIN_FOOT_CMDS

