#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "camera.h"
#include "config.h"
#include "edsel.h"
#include "entity.h"
#include "geom.h"
#include "map.h"
#include "quest.h"
#include "s_3d.h"
#include "s_error.h"
#include "s_message.h"
#include "status.h"
#include "transform.h"
#include "uid.h"


typedef struct
{
   vec3_t base;

   brush_t *BrushHead;

   entity_t *EntityHead;
   brush_t *EntityBrush;
} clipboard_t;

static clipboard_t C;


static void FreeC(void)
{
   brush_t *b,*bn;
   entity_t *e,*en;

   for (b=C.BrushHead;b;b=bn)
   {
      bn=b->Next;
      B_Free(b);
   }
   C.BrushHead=NULL;
   
   for (b=C.EntityBrush;b;b=bn)
   {
      bn=b->Next;
      B_Free(b);
   }
   C.EntityBrush=NULL;
   
   for (e=C.EntityHead;e;e=en)
   {
      en=e->Next;
      E_Free(e);
   }
   C.EntityHead=NULL;
}


static void C_copy(void)
{
   bsel_t *bs;
   esel_t *es;
   brush_t *b,*b2;
   entity_t *e,*e2;
   int num;

   if (!M.sel.esel && !M.sel.bsel)
   {
      S_NewMessage("edit.copy: Nothing selected!");
      return;
   }
   
   FreeC();
   C.base=S_cur_vport->camera_pos;

   num=1;
   for (es=M.sel.esel;es;es=es->Next)
   {
      e=es->Entity;
      e->misc=num++;

      e2=E_Duplicate(NULL,e);
      if (!e2)
      {
         FreeC();
         S_Error("edit.copy","Out of memory!");
         return;
      }
      e2->uid=e->misc;
      e2->Next=C.EntityHead;
      if (C.EntityHead) C.EntityHead->Last=e2;
      C.EntityHead=e2;
   }

   for (b=M.VBrushHead;b;b=b->VNext)
      if ((e=UID_Entity(b->Entity)))
         if (e->misc)
         {
            b->misc=1;

            b2=B_Duplicate(NULL,b);
            if (!b2)
            {
               FreeC();
               S_Error("edit.copy","Out of memory!");
               return;
            }
            b2->Entity=e->misc;
            b2->Next=C.EntityBrush;
            if (C.EntityBrush) C.EntityBrush->Last=b2;
            C.EntityBrush=b2;
         }
            
   for (bs=M.sel.bsel;bs;bs=bs->Next)
   {
      b=bs->Brush;
      if (b->misc) continue;

      b2=B_Duplicate(NULL,b);
      if (!b2)
      {
         FreeC();
         S_Error("edit.copy","Out of memory!");
         return;
      }
      b2->Next=C.BrushHead;
      if (C.BrushHead) C.BrushHead->Last=b2;
      C.BrushHead=b2;
   }
   
   for (b=M.VBrushHead;b;b=b->VNext)
      b->misc=0;
   
   for (es=M.sel.esel;es;es=es->Next)
      es->Entity->misc=0;
}


static void C_paste(void)
{
   vec3_t delta;
   entity_t *e,*e2;
   brush_t *b,*b2;

   if (!C.EntityHead && !C.BrushHead)
   {
      S_NewMessage("edit.paste: Nothing to paste!");
      return;
   }

   ClearSelAll();

   VEC_Sub(delta,S_cur_vport->camera_pos,C.base);
   delta.x=SnapPointToGrid(delta.x);
   delta.y=SnapPointToGrid(delta.y);
   delta.z=SnapPointToGrid(delta.z);
   if (!delta.x && !delta.y && !delta.z)
   {
      Move90(S_cur_vport,MOVE_UP,&delta,cv_edit_snap_size->i);
   }

   for (e=C.EntityHead;e;e=e->Next)
   {
      e2=E_Duplicate(NULL,e);
      if (!e2)
      {
         S_Error("edit.paste","Out of memory!");
         return;
      }
      E_Link(e2);

      for (b=C.EntityBrush;b;b=b->Next)
      {
         if (b->Entity!=e->uid) continue;
         b2=B_Duplicate(NULL,b);
         if (!b2)
         {
            S_Error("edit.paste","Out of memory!");
            return;
         }
         b2->Entity=e2->uid;
         B_Link(b2);
         AddSelBrush(b2);
      }
      AddSelEnt(e2);
   }

   for (b=C.BrushHead;b;b=b->Next)
   {
      b2=B_Duplicate(NULL,b);
      if (!b2)
      {
         S_Error("edit.paste","Out of memory!");
         return;
      }
      B_Link(b2);
      AddSelBrush(b2);
   }
   TransformMove(delta,&M.sel);
   Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"edit.copy" ,C_copy,{"Edit/Copy","control+c"},{},"Copies the current selection to the clipboard."},
{"edit.paste",C_paste,{"Edit/Paste","control+v"},{},"Pastes the contents of the clipboard."},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/edit/clipboard","Clipboard commands."}};

PLUGIN_FOOT_CMDS

