#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "config.h"
#include "edsel.h"
#include "entity.h"
#include "quest.h"
#include "s_message.h"
#include "s_gui.h"
#include "status.h"


static void C_grid_inc(void)
{
   Cfg_SetInt(cv_edit_snap_size,cv_edit_snap_size->i*2);
   S_NewMessage("Snap size: %i",cv_edit_snap_size->i);
   Cmd_SetRedraw(2);
}

static void C_grid_dec(void)
{
   Cfg_SetInt(cv_edit_snap_size,cv_edit_snap_size->i/2);
   if (cv_edit_snap_size->i < 1) Cfg_SetInt(cv_edit_snap_size,1);
   S_NewMessage("Snap size: %i",cv_edit_snap_size->i);
   Cmd_SetRedraw(2);
}


static void C_delete(void)
{
   bsel_t *b;
   esel_t *e;

   if (!M.sel.bsel && !M.sel.esel)
   {
      S_NewMessage("edit.delete: Nothing selected.");
      return;
   }

   for (b=M.sel.bsel;b;b=b->Next)
   {
      B_Unlink(b->Brush);
      B_Free(b->Brush);
   }
   
   for (e=M.sel.esel;e;e=e->Next)
   {
      if (e->Entity->uid==M.WorldSpawn)
         if (!S_GUI_YesNo("Delete worldspawn?",
            "Warning: You are about to delete the worldspawn entity. "
            "Are you REALLY sure you want to do this?"))
            continue;
   
      E_Unlink(e->Entity);
      E_Free(e->Entity);
   }

   ClearSelAll();
   Cmd_SetRedraw(2);
}


static void C_texlock(void)
{
   c_texture_locking=!c_texture_locking;
   S_NewMessage("Texture locking %sabled.",
      c_texture_locking?"en":"dis");
}


static cmd_reg_t cmds[]={
{"edit.grid_inc",C_grid_inc,{NULL,"plus" },{},"Increases the size of the grid."},
{"edit.grid_dec",C_grid_dec,{NULL,"minus"},{},"Decreases the size of the grid."},

{"edit.delete",C_delete,{"Edit/Delete","delete"},{},"Deletes all selected brushes and entities."},

{"edit.texlock",C_texlock,{"Edit/Texture locking","e"},{},"Toggle texture locking."},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/edit/misc","Misc. fundamental editing commands."}};

PLUGIN_FOOT_CMDS

