#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "camera.h"
#include "config.h"
#include "quest.h"
#include "s_3d.h"
#include "s_message.h"
#include "status.h"
#include "transform.h"


static void MoveSel(int dir)
{
   vec3_t delta;

   if (!M.sel.vsel && !M.sel.esel && !M.sel.bsel && !M.sel.fsel)
   {
      S_NewMessage("edit.move: Nothing selected!");
      return;
   }

   Move90(S_cur_vport,dir,&delta,cv_edit_snap_size->i);

   TransformMove(delta,&M.sel);

   Cmd_SetRedraw(2);
}


static void C_move_left    (void) { MoveSel(MOVE_LEFT    ); }
static void C_move_right   (void) { MoveSel(MOVE_RIGHT   ); }
static void C_move_up      (void) { MoveSel(MOVE_UP      ); }
static void C_move_down    (void) { MoveSel(MOVE_DOWN    ); }
static void C_move_forward (void) { MoveSel(MOVE_FORWARD ); }
static void C_move_backward(void) { MoveSel(MOVE_BACKWARD); }

static const char help[]=
"The edit.move_xxx commands move the current selection in different "
"directions. The distance moved is controlled by edit.snap_size. "
"\n\n"
"If any vertices are selected, they will be moved. If no vertices are "
"selected, all vertices of all selected brushes and faces are moved. "
"If any entities are selected, they will also be moved. If they have "
"an origin field, it will be updated. If an entity has any brushes attached, "
"they will also be moved. If a brush is part of a selected entity, and is "
"selected itself, it will only be moved once.";

static cmd_reg_t cmds[]={
{"edit.move_left"    ,C_move_left    ,{NULL,"shift+left"     },{0,"edit.move"},"Moves selection left."     ,help},
{"edit.move_right"   ,C_move_right   ,{NULL,"shift+right"    },{0,"edit.move"},"Moves selection right."    ,help},
{"edit.move_up"      ,C_move_up      ,{NULL,"shift+up"       },{0,"edit.move"},"Moves selection up."       ,help},
{"edit.move_down"    ,C_move_down    ,{NULL,"shift+down"     },{0,"edit.move"},"Moves selection down."     ,help},
{"edit.move_forward" ,C_move_forward ,{NULL,"shift+page-up"  },{0,"edit.move"},"Moves selection forwards." ,help},
{"edit.move_backward",C_move_backward,{NULL,"shift+page-down"},{0,"edit.move"},"Moves selection backwards.",help},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/edit/move","Fundamental selection movement commands."}};

PLUGIN_FOOT_CMDS

