#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "camera.h"
#include "edsel.h"
#include "entclass.h"
#include "entity.h"
#include "geom.h"
#include "map.h"
#include "memory.h"
#include "quest.h"
#include "s_3d.h"
#include "s_error.h"
#include "s_gui.h"
#include "s_message.h"
#include "uid.h"
#include "undo.h"


static void C_add_brush(void)
{
	bsel_t *b;
	entity_t *e;

	if (!M.sel.num_bselected)
	{
		S_NewMessage("entity.add_brush: At least one brush must be selected.");
		return;
	}

   if (M.sel.num_eselected!=1)
   {
		S_NewMessage("entity.add_brush: Exactly one entity must be selected.");
		return;
	}

   e=M.sel.esel->Entity;
	for (b=M.sel.bsel;b;b=b->Next)
   {
      U_BrushModify(b->Brush);
		b->Brush->Entity = e->uid;
   }

	Cmd_SetRedraw(2);
}


static void C_create(void)
{
   char buf[256];
   entity_t *e;
   vec3_t d;

   if (!S_GUI_EntityPicker(buf,ECLASS_POINT))
      return;

   e=E_New();
   if (!e)
   {
      S_Error("entity.create","Out of memory!");
      return;
   }
   SetKeyValue(e,"classname",buf);
   
   Move(S_cur_vport,MOVE_FORWARD,&d,100); /* TODO: edit.project_object */
   VEC_Add(d,d,S_cur_vport->camera_pos);
   sprintf(buf,"%i %i %i",(int)d.x,(int)d.y,(int)d.z);
   SnapVecToGrid(&d);
   SetKeyValue(e,"origin",buf);

   SetEntityDefaults(e);
   E_Link(e);
   
   ClearSelAll();
   AddSelEnt(e);
   Cmd_SetRedraw(2);
}


static void C_create_model(void)
{
   char classname[256];
   entity_t *e;
	bsel_t *bs;

	if (M.sel.num_bselected == 0)
	{
		S_Error("entity.create_model", "At least one brush must be selected.");
		return;
	}
   if (!S_GUI_EntityPicker(classname,ECLASS_MODEL))
      return;

   e=E_New();
   if (!e)
   {
      S_Error("entity.create_model","Out of memory!");
      return;
   }

   SetKeyValue(e, "classname", classname);
   SetEntityDefaults(e);

   E_Link(e);
   
   /* Assign the brushes to the entity */
   for (bs=M.sel.bsel; bs; bs=bs->Next)
   {
      U_BrushModify(bs->Brush);
   	bs->Brush->Entity = e->uid;
   }
   
   ClearSelBrushes();
   Cmd_SetRedraw(2);
}


static void C_edit(void)
{
   entity_t **ents;
   int nents;
   esel_t *e;
   bsel_t *bs;
   map_uid_t eu;
   int i;

	if (!M.sel.esel && !M.sel.bsel)
   {
      S_NewMessage("entity.edit: Nothing selected!");
      return;
   }

   nents=0;
   ents=NULL;
   for (e=M.sel.esel;e;e=e->Next)
   {
      nents++;
      ents=Q_realloc(ents,sizeof(entity_t *)*nents);
      if (!ents)
      {
         S_Error("entity.edit","Out of memory!");
         return;
      }
      ents[nents-1]=e->Entity;
   }
   for (bs=M.sel.bsel;bs;bs=bs->Next)
   {
      eu=bs->Brush->Entity;
      for (i=0;i<nents;i++)
         if (eu==ents[i]->uid)
            break;
      if (i!=nents) continue;
      nents++;
      ents=Q_realloc(ents,sizeof(entity_t *)*nents);
      if (!ents)
      {
         S_Error("entity.edit","Out of memory!");
         return;
      }
      ents[nents-1]=UID_Entity(eu);
   }
   
   E_Edit(nents,ents);

   Q_free(ents);
   Cmd_SetRedraw(2);
}



static cmd_reg_t cmds[]={
{"entity.add_brush",C_add_brush,{"Entity/Add brush"},{},"Add the selected brushes to the selected entity."},
{"entity.create",C_create,{"Entity/Create...","alt+e"},{},"Create an entity."},
{"entity.create_model",C_create_model,{"Entity/Create model..."},{},"Create a model."},
{"entity.edit",C_edit,{"Entity/Edit...","control+e"},{},"Edit the selected entities and the entities of the selected brushes."},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/entity","Basic entity commands."}};

PLUGIN_FOOT_CMDS

