#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "edsel.h"
#include "entity.h"
#include "quest.h"
#include "s_error.h"
#include "s_int.h"
#include "s_message.h"
#include "s_render.h"


static void CreateLink(sel_t *from,sel_t *to,const char *src_name,const char *dst_name)
{
   esel_t *es;
   char targetname[256];
   const char *c;

	/* Check if they're trying to link from worldspawn */
	for (es=from->esel; es; es=es->Next)
		if (es->Entity->uid==M.WorldSpawn)
			S_NewMessage("entity.link: Linking from worldspawn!");

   /* Check if they're trying to link to worldspawn */
	for (es=to->esel; es; es=es->Next)
		if (es->Entity->uid==M.WorldSpawn)
			S_NewMessage("entity.link: Linking to worldspawn!");

   targetname[0]=0;
   for (es=from->esel;es;es=es->Next)
   {
      if ((c=GetKeyValue(es->Entity,src_name)))
      {
         if (targetname[0])
         {
            if (strcmp(targetname,c))
            {
               S_Error("entity.link","Entities in 'from' list have different '%s' keys!",src_name);
         		return;
            }
         }
         else
            strcpy(targetname,c);
      }
   }
   for (es=to->esel;es;es=es->Next)
   {
      if ((c=GetKeyValue(es->Entity,dst_name)))
      {
         if (targetname[0])
         {
            if (strcmp(targetname,c))
            {
               S_Error("entity.link","Entities in 'to' list have different '%s' keys!",dst_name);
         		return;
            }
         }
         else
            strcpy(targetname,c);
      }
   }

   if (!targetname[0])
   {
      int i,j;

      /* TODO: this may not always work. E.g. for target/targetname, we might
      use a t%i string that some entity has as a pathtarget key (which would
      be a broken link, but still).
      Idea: A 'edit.link_names' cvar which lists all of them, and a function
      in entity.c that checks all of them. Plugins would be free to add stuff
      to the list.
      */
      i=Target_GetLastNum(src_name);
      j=Target_GetLastNum(dst_name);
      if (i<j) i=j;
      i++;
      sprintf(targetname,"t%i",i);
   }

	for (es=from->esel;es;es=es->Next)
		SetKeyValue(es->Entity,src_name,targetname);
	for (es=to->esel;es;es=es->Next)
		SetKeyValue(es->Entity,dst_name,targetname);

	S_NewMessage("Link Created (using '%s').",targetname);
}


static void C_link(void)
{
   sel_t from;

   if (!M.sel.num_eselected)
   {
      S_NewMessage("entity.link: No entities selected!");
      return;
   }

   from=M.sel;
   memset(&M.sel,0,sizeof(sel_t));
   M.sel.flags=SELECTION_PRIMARY;
   S_R_SelUpdate();

   S_I_GetSelection("Select the targets.");
   
	if (!M.sel.num_eselected)
	{
      ClearSelAll();
      M.sel=from;
		S_NewMessage("entity.linl: No targets selected.");
		return;
	}

   CreateLink(&from,&M.sel,"target","targetname");

   ClearSelAll();
   M.sel=from;

	Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"entity.link",C_link,{"Entity/Link","control+l"},{},"Create a link between entities."},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/link","Create links between entities."}};

PLUGIN_FOOT_CMDS

