#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "config.h"
#include "edsel.h"
#include "quest.h"
#include "status.h"
#include "s_message.h"


static int current_mode;
#define M_BRUSH  0
#define M_FACE   1
#define M_ENTITY 2
#define M_MODEL  3
static int rflags[4]=
{
   RENDER_BRUSH|
   RENDER_BRUSH_FOCAL|
   RENDER_BRUSH_VERT
,
   RENDER_BRUSH|
   RENDER_FACE_FOCAL|
   RENDER_FACE_VERT
,
   RENDER_BRUSH|
   RENDER_ENTITY|
   RENDER_ENTITY_FOCAL|
   RENDER_ENTITY_LINK
,
   RENDER_BRUSH|
   RENDER_BRUSH_FOCAL|
   RENDER_BRUSH_VERT|
   RENDER_ENTITY|
   RENDER_ENTITY_FOCAL|
   RENDER_ENTITY_LINK
};

static const char *mname[4]={"brush","face","entity","model"};

static void SetMode(int m)
{
   current_mode=m;
   Cfg_SetInt(cv_render_flags,rflags[m]);
   switch (m)
   {
   case M_BRUSH:
      ClearSelFaces();
      ClearSelEnts();
      break;
   case M_FACE:
      ClearSelBrushes();
      ClearSelEnts();
      break;
   case M_ENTITY:
      ClearSelVerts();
      ClearSelBrushes();
      ClearSelFaces();
      break;
   case M_MODEL:
      ClearSelVerts();
      ClearSelFaces();
      break;
   }
   S_NewMessage("Switched to %s mode.",mname[m]);
   Cmd_SetRedraw(2);
}


static void C_next(void)
{
   SetMode((current_mode+1)%4);
}

static void C_prev(void)
{
   SetMode((current_mode+3)%4);
}

static void C_m_brush (void) { SetMode(M_BRUSH ); }
static void C_m_face  (void) { SetMode(M_FACE  ); }
static void C_m_entity(void) { SetMode(M_ENTITY); }
static void C_m_model (void) { SetMode(M_MODEL ); }


static cmd_reg_t mode_cmds[]={
{"mode.next",C_next,{"View/Mode/Next","tab"      },{},"Switches to the next mode."    },
{"mode.prev",C_prev,{"View/Mode/Prev","shift+tab"},{},"Switches to the previous mode."},

{"mode.brush" ,C_m_brush ,{"View/Mode/Brush" },{},"Switches to brush mode." },
{"mode.face"  ,C_m_face  ,{"View/Mode/Face"  },{},"Switches to face mode."  },
{"mode.entity",C_m_entity,{"View/Mode/Entity"},{},"Switches to entity mode."},
{"mode.model" ,C_m_model ,{"View/Mode/Model" },{},"Switches to model mode." },
};

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/modes","Emulates old mode system."}};

static int Init(void)
{
   Cmd_Register(&plugin,mode_cmds,sizeof(mode_cmds));
   SetMode(M_BRUSH);
   return 1;
}

PLUGIN_FOOT

