#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "edsel.h"
#include "geom.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"
#include "s_int.h"
#include "s_message.h"
#include "s_render.h"
#include "undo.h"


static void C_snap_to_vertex(void)
{
	vsel_t *vlist;
	bsel_t *blist;
	vsel_t *v;
	bsel_t *b;
	int     numv,numb;
	vsel_t *vhandle;
	vec3_t  diff;

   if (!M.sel.num_vselected)
   {
      S_Error("brush.snap_to_vertex",
         "Must have at least one vertex selected!");
      return;
   }

	vlist = M.sel.vsel;			/* Remember selected verts */
	numv = M.sel.num_vselected;
	M.sel.num_vselected = 0;
	M.sel.vsel = NULL;

	blist = M.sel.bsel;			/* Remember selected brushes */
	numb = M.sel.num_bselected;
	M.sel.num_bselected = 0;
	M.sel.bsel = NULL;

   S_R_SelUpdate();
   
	/* Now that we have the list of vertices to move, we need one to
	serve as the "handle" to drag from */

	if (numv > 1)
	{
		/* re-add the brushes that were selected to the display.list
      because we usually want to choose one vertex from them */
      
		for (b=blist; b; b=b->Next)
		{
			AddSelBrush(b->Brush);
		}

      S_I_GetSelection("Select vertex to drag from.");

		if (M.sel.num_vselected == 1)
		{
			vhandle = M.sel.vsel;			/* Remember it */
			M.sel.num_vselected = 0;
			M.sel.vsel = NULL;
		}
		else
		{
         ClearSelVerts();
			ClearSelBrushes();
			M.sel.num_vselected = numv;	/* Put original selection back */
			M.sel.vsel = vlist;
			M.sel.num_bselected = numb;
			M.sel.bsel = blist;
         S_R_SelUpdate();
			S_NewMessage("brush.snap_to_vertex: No handle vertex selected. Aborted.");
			return;
		}
	}
	else
	{
		/* Use the single selected vertex as the handle vertex */
		vhandle = vlist;
	}

	ClearSelBrushes();
   S_I_GetSelection("Select vertex to snap to.");

	if (M.sel.num_vselected != 1)
	{
		if (numv > 1)
		{
			Q_free(vhandle);			/* Free handle vertex */
		}
		ClearSelBrushes();
		M.sel.num_vselected = numv;	/* Put original selection back */
		M.sel.vsel = vlist;
		M.sel.num_bselected = numb;
		M.sel.bsel = blist;
      S_R_SelUpdate();
		S_NewMessage("brush.snap_to_vertex: No snap to vertex selected. Aborted.");
      Cmd_SetRedraw(2);
		return;
	}

	/* Compute amount to add to selected verts to snap them */
   VEC_Sub(diff,M.sel.vsel->Brush->verts[M.sel.vsel->num],vhandle->Brush->verts[vhandle->num]);

	ClearSelBrushes();
	M.sel.num_bselected = numb;
   M.sel.bsel = blist;

	/* Modify originally selected vertices by the computed amount */
	for (v=vlist; v; v=v->Next)
   {
      U_BrushModify(v->Brush);
   
      VEC_Add(v->Brush->verts[v->num],v->Brush->verts[v->num],diff);
	}

	/* Perform cleanup */
	if (numv > 1)
	{
		Q_free(vhandle);	/* Free handle vertex selected earlier */
	}

	/* Leave originally selected brushes selected so user can see the changes */
	M.sel.vsel = vlist;			/* Remember selected verts */
	M.sel.num_vselected = numv;
	M.sel.bsel = blist;			/* Remember selected brushes */
	M.sel.num_bselected = numb;

   Cmd_SetRedraw(2);
}


static cmd_reg_t cmds[]={
{"brush.snap_to_vertex",C_snap_to_vertex,{"Brush/Snap to vertex","u"},{},"Snaps selected vertices to a vertex.",
/* TODO */
"This function is used to translate one or more selected vertices "
"along a plane to another vertex.  It can be used to precisely join "
"brush corners and edges when no brush vertices are snapped to a grid.\n"
"\n"
"The selected vertices are first copied off to a list of vertices "
"that will be modified.\n"
"\n"
"The user is then asked to select a single handle vertex for the group.\n"
"\n"
"Finally, the user selects the destination vertex.  The amount to translate "
"the original set by is the difference between the handle and the "
"destination.\n"
"\n"
"---\n"
"originally by Gyro Gearloose\n"
},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/snap_to_vertex","Command to snap to a vertex."}};

PLUGIN_FOOT_CMDS

