#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "types.h"

#include "s_texture.h"
#include "gl_texture.h"

#include "memory.h"
#include "s_error.h"


static tex_funcs_t *tf;


static gl_texture_t **tex;
static int num_tex;

int S_tex_use_cache;


void S_GL_GetTexs(gl_texture_t ***ptex,int *pnum_tex)
{
   *ptex=tex;
   *pnum_tex=num_tex;
}



static unsigned char palette[768];


int S_GL_WantLoad(gl_texture_t *t,int level)
{
   if (level<=t->load) return 1;

   if (t->load==-1)
      return 0;

   if (!tf)
      return 0;

   if (level==1 && tf->load_size)
   {
      tf->load_size(t,&t->tex);
      if (t->load==1)
         return 1;
   }

   if (!tf->load)
      return 0;

   tf->load(t,&t->tex);
   if (t->load<level)
   {
      t->load=-1;
      return 0;
   }
   return 1;
}


static gl_texture_t *FindTexture(const char *name)
{
   int l,h,m;
   int i;

   if (!num_tex) return NULL;

   for (l=0,h=num_tex-1;l<h;)
   {
      m=(l+h)/2;
      if (!(i=stricmp(name,tex[m]->tex.name)))
         return tex[m];

      if (i<0) h=m-1; else l=m+1;
   }
   if (!stricmp(name,tex[l]->tex.name))
      return tex[l];
   return NULL;
}


gl_texture_t *S_GL_LoadTexture(const char *name)
{
   gl_texture_t *t;
   t=FindTexture(name);
   if (!t) return NULL;
   if (t->load>=2) return t;
   if (S_GL_WantLoad(t,2))
      return t;
   return NULL;
}


static gl_texture_t *InsertTexture(const char *name,gl_texture_t *t)
{
   int l,h,m;
   int i;

   gl_texture_t **tt;

   if (num_tex)
   {
      for (l=0,h=num_tex-1;l<h;)
      {
         m=(l+h)/2;
         if (!(i=stricmp(name,tex[m]->tex.name)))
            return NULL;

         if (i<0) h=m-1; else l=m+1;
      }
      i=stricmp(name,tex[l]->tex.name);
      if (!i) return 0;
      if (i>0) l++;
   }
   else
      l=0;

   tt=Q_realloc(tex,sizeof(gl_texture_t *)*(num_tex+1));
   if (!tt) return NULL;
   tex=tt;

   for (h=num_tex;h>l;h--)
      tex[h]=tex[h-1];
   tex[h]=t;
   num_tex++;

   return t;
}


texture_t *S_T_Load(const char *name)
{
   gl_texture_t *t;

   t=FindTexture(name);
   if (!t)
      return NULL;
   t->tex.cached=1;
   if (!S_GL_WantLoad(t,2))
      return NULL;
   return &t->tex;
}

texture_t *S_T_LoadSize(const char *name)
{
   gl_texture_t *t;

   t=FindTexture(name);
   if (!t)
      return NULL;
   t->tex.cached=1;
   if (!S_GL_WantLoad(t,1))
      return NULL;
   return &t->tex;
}

int S_T_LoadCheck(const char *name)
{
   if (FindTexture(name))
      return 1;
   return 0;
}


void S_T_GetDescription(char *dst,const texture_t *t)
{
   if (!tf->tex_desc)
   {
      *dst=0;
      return;
   }
   tf->tex_desc(dst,t);
}


int S_T_Cache(const char *name)
{
   gl_texture_t *t;
   if (!S_tex_use_cache) return 1;
   t=FindTexture(name);
   if (!t)
      return 0;
   t->tex.cached=1;
   return 1;
}

int S_T_Uncache(const char *name)
{
   gl_texture_t *t;
   if (!S_tex_use_cache) return 1;
   t=FindTexture(name);
   if (!t)
      return 0;
   t->tex.cached=0;
   return 1;
}


void S_T_Init(int cache,tex_funcs_t *ptf)
{
   S_tex_use_cache=cache;
   S_T_Clear();
   memset(palette,0,sizeof(palette));
   tf=ptf;
}

void S_T_Clear(void)
{
   int i;
   tf=NULL;
   for (i=0;i<num_tex;i++)
   {
      if (tex[i]->load==2)
         glDeleteTextures(1,&(tex[i]->num));
      if (tex[i]->tex.g)
         Q_free(tex[i]->tex.g);
      Q_free(tex[i]);
   }
   Q_free(tex);
   tex=NULL;
   num_tex=0;
}

void S_GL_Texture_Done(void)
{
   S_T_Clear();
}


void S_T_ClearCache(void)
{
   int i;
   if (!S_tex_use_cache) return;
   for (i=0;i<num_tex;i++)
      tex[i]->tex.cached=0;
}


void S_T_SetPalette(unsigned char *pal)
{
   memcpy(palette,pal,sizeof(palette));

#if 1
/* Some simple gamma adjustments */
   { /* TODO */
      float f;
      int i;

      for (i=0;i<768;i++)
      {
         f=pow(palette[i]/255.0,0.85)*255.0;
         if (f<0) f=0;
         if (f>255) f=255;
         palette[i]=f;
      }
   }
#endif
}


texture_t *S_T_GetTexture(void *tex)
{
   gl_texture_t *t=tex;
   return &t->tex;
}


void *S_T_AddTexture(const char *name,int width,int height,void *g)
{
   gl_texture_t *t;

   if (FindTexture(name))
      return 0;

   t=Q_malloc(sizeof(gl_texture_t));
   if (!t)
      return 0;
   memset(t,0,sizeof(gl_texture_t));
   strcpy(t->tex.name,name);
   t->tex.width=width;
   t->tex.height=height;
   t->tex.g=g;

   if (!S_tex_use_cache)
      t->tex.cached=1;

   if (width==-1)
      t->load=0;
   else
      t->load=1;

   return InsertTexture(name,t);
}


int S_T_AddSize(void *tex,int width,int height)
{
   gl_texture_t *t=tex;

   if (t->load>=1) return 0;
   if (width==-1) return 0;
   t->tex.width=width;
   t->tex.height=height;
   t->load=1;
   return 1;
}

int S_T_AddData8(void *tex,int width,int height,int data_x,int data_y,unsigned char *data,unsigned char *ppal)
{
   unsigned char *pal;
   unsigned char *buf;
   int i;
   unsigned char *c,*d;

   if (ppal)
      pal=ppal;
   else
      pal=palette;

   if (!data_x)
   {
      data_x=width;
      data_y=height;
   }

   buf=Q_malloc(3*data_x*data_y);
   if (!buf)
      return 0;

   for (i=data_x*data_y,c=data,d=buf;i;i--)
   {
      *d++=pal[*c *3+0];
      *d++=pal[*c *3+1];
      *d++=pal[*c *3+2];
      c++;
   }

   i=S_T_AddData24(tex,width,height,data_x,data_y,buf);
   Q_free(buf);
   return i;
}

int S_T_AddData24(void *tex,int width,int height,int data_x,int data_y,unsigned char *data)
{
   gl_texture_t *t=tex;

   t->tex.width=width;
   t->tex.height=height;

   /* TODO: fix bug in mesa/glx/nvidia that causes hard x crash on shutdown if
   gluBuild2DMipmaps has been called with size==0 */
   if (!data_x)
   {
      data_x=width;
      data_y=height;
   }

   glGenTextures(1,&t->num);

   glBindTexture(GL_TEXTURE_2D,t->num);
   gluBuild2DMipmaps(GL_TEXTURE_2D,3,data_x,data_y,GL_RGB,GL_UNSIGNED_BYTE,data);
//   glTexImage2D(GL_TEXTURE_2D,0,3,width,height,0,GL_RGB,GL_UNSIGNED_BYTE,data);
      
/*   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);*/
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
   glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
      
   glBindTexture(GL_TEXTURE_2D,0);

   t->load=2;

   return 1;
}

