#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gtk/gtk.h>

#include "s_error.h"

#include "game.h"
#include "g_gui.h"


static int in_abort=0;


void S_Error(const char *proc, const char *format, ...)
{
   char buf[256];
   va_list args;

   if (in_abort) return;

   va_start(args,format);
   vsnprintf(buf,sizeof(buf)-1,format,args);
   va_end(args);

   S_GUI_DialogOK("Error!","%s:\n%s",proc,buf);
}


/*  Like HandleError, but will terminate the program after attempting to
save the current map.
*/
void S_Abort(const char *proc,const char *format, ...)
{
   va_list args;

   if (!in_abort)
   { /* Try to save the map, first to in the current directory, then in
     the home directory, then in /tmp */
      const char **n;
static const char *backup_names[]={
"./q_backup.map",
"~/q_backup.map",
"/tmp/q_backup.map",
0
};
      in_abort=1;

      for (n=backup_names;*n;n++)
         if (G_SaveMap(*n))
            break;
      if (*n)
         fprintf(stderr,"Saved current map in '%s'.\n",*n);
      else
         fprintf(stderr,"Tried to save current map but failed!\n");
   }
   else
   {
      fprintf(stderr,"Recursive abort! Error saving map?\n");
   }

   fprintf(stderr,"%s: ",proc);
   va_start(args,format);
   vfprintf(stderr,format,args);
   va_end(args);
   fprintf(stderr,"\nProgram aborted!\n");
   gtk_exit(1);
   exit(1);
   abort();
}

