#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "g_input.h"
#include "s_input.h"

#include "gl/g_3d.h"

#include "cmds.h"
#include "s_error.h"


/* TODO: think of some better way to do this */
#define ARG_UNUSED(x) x __attribute__ ((unused))


typedef struct
{
   int gdk;   /* GDK's value for the key */
   int quest; /* Quest's value to convert to */
   int mod;   /* Set to 1 if this key can safely be combined with modifiers
                 (creative keymaps might break this, but it's better than
                 nothing. */
} conv_t;

static conv_t conv[]={
{GDK_Escape        ,KEY_ESCAPE   ,1},
{GDK_Return        ,KEY_ENTER    ,1},
{GDK_KP_Enter      ,KEY_ENTER    ,1},
{GDK_space         ,KEY_SPACE    ,1},

{GDK_Tab           ,KEY_TAB      ,1},
{GDK_ISO_Left_Tab  ,KEY_TAB      ,1},

{GDK_Up            ,KEY_UP       ,1},
{GDK_Down          ,KEY_DOWN     ,1},
{GDK_Left          ,KEY_LEFT     ,1},
{GDK_Right         ,KEY_RIGHT    ,1},

{GDK_BackSpace     ,KEY_BACKSPC  },
#if TODO /* TODO: do these properly, will have to fix s_input.h and s_input.c too */
{GDK               ,KEY_COLON    },
{GDK               ,KEY_QUOTE    },
{GDK               ,KEY_TILDE    },
{GDK               ,KEY_PIPE     },
{GDK               ,KEY_COMMA    },
{GDK               ,KEY_PERIOD   },
{GDK               ,KEY_Q_MARK   },
{GDK               ,KEY_LF_BRACK },
{GDK               ,KEY_RT_BRACK },
{GDK               ,KEY_CENTER   },
#endif
{GDK_Insert        ,KEY_INSERT   ,1},
{GDK_Delete        ,KEY_DELETE   ,1},
{GDK_Home          ,KEY_HOME     ,1},
{GDK_End           ,KEY_END      ,1},
{GDK_Page_Up       ,KEY_PAGEUP   ,1},
{GDK_Page_Down     ,KEY_PAGEDOWN ,1},
{GDK_KP_Subtract   ,KEY_MINUS    ,1},
{GDK_KP_Add        ,KEY_PLUS     ,1},

{GDK_A,KEY_A,1},{GDK_a,KEY_A,1},
{GDK_B,KEY_B,1},{GDK_b,KEY_B,1},
{GDK_C,KEY_C,1},{GDK_c,KEY_C,1},
{GDK_D,KEY_D,1},{GDK_d,KEY_D,1},
{GDK_E,KEY_E,1},{GDK_e,KEY_E,1},
{GDK_F,KEY_F,1},{GDK_f,KEY_F,1},
{GDK_G,KEY_G,1},{GDK_g,KEY_G,1},
{GDK_H,KEY_H,1},{GDK_h,KEY_H,1},
{GDK_I,KEY_I,1},{GDK_i,KEY_I,1},
{GDK_J,KEY_J,1},{GDK_j,KEY_J,1},
{GDK_K,KEY_K,1},{GDK_k,KEY_K,1},
{GDK_L,KEY_L,1},{GDK_l,KEY_L,1},
{GDK_M,KEY_M,1},{GDK_m,KEY_M,1},
{GDK_N,KEY_N,1},{GDK_n,KEY_N,1},
{GDK_O,KEY_O,1},{GDK_o,KEY_O,1},
{GDK_P,KEY_P,1},{GDK_p,KEY_P,1},
{GDK_Q,KEY_Q,1},{GDK_q,KEY_Q,1},
{GDK_R,KEY_R,1},{GDK_r,KEY_R,1},
{GDK_S,KEY_S,1},{GDK_s,KEY_S,1},
{GDK_T,KEY_T,1},{GDK_t,KEY_T,1},
{GDK_U,KEY_U,1},{GDK_u,KEY_U,1},
{GDK_V,KEY_V,1},{GDK_v,KEY_V,1},
{GDK_W,KEY_W,1},{GDK_w,KEY_W,1},
{GDK_X,KEY_X,1},{GDK_x,KEY_X,1},
{GDK_Y,KEY_Y,1},{GDK_y,KEY_Y,1},
{GDK_Z,KEY_Z,1},{GDK_z,KEY_Z,1},

{GDK_0,KEY_0,0},{GDK_KP_0,KEY_0,1},
{GDK_1,KEY_1,0},{GDK_KP_1,KEY_1,1},
{GDK_2,KEY_2,0},{GDK_KP_2,KEY_2,1},
{GDK_3,KEY_3,0},{GDK_KP_3,KEY_3,1},
{GDK_4,KEY_4,0},{GDK_KP_4,KEY_4,1},
{GDK_5,KEY_5,0},{GDK_KP_5,KEY_5,1},
{GDK_6,KEY_6,0},{GDK_KP_6,KEY_6,1},
{GDK_7,KEY_7,0},{GDK_KP_7,KEY_7,1},
{GDK_8,KEY_8,0},{GDK_KP_8,KEY_8,1},
{GDK_9,KEY_9,0},{GDK_KP_9,KEY_9,1},

{GDK_F1 ,KEY_F1 ,1},
{GDK_F2 ,KEY_F2 ,1},
{GDK_F3 ,KEY_F3 ,1},
{GDK_F4 ,KEY_F4 ,1},
{GDK_F5 ,KEY_F5 ,1},
{GDK_F6 ,KEY_F6 ,1},
{GDK_F7 ,KEY_F7 ,1},
{GDK_F8 ,KEY_F8 ,1},
{GDK_F9 ,KEY_F9 ,1},
{GDK_F10,KEY_F10,1},
{GDK_F11,KEY_F11,1},
{GDK_F12,KEY_F12,1}
};
static const int num_conv=sizeof(conv)/sizeof(conv[0]);


int S_Input_ConvState(int s)
{
   int i=0;
   if (s&GDK_SHIFT_MASK) i|=KEYM_SHIFT;
   if (s&GDK_MOD1_MASK) i|=KEYM_ALT;
   if (s&GDK_CONTROL_MASK) i|=KEYM_CONTROL;
   return i;
}

int S_Input_ConvKey(int k,int s)
{
   conv_t *c;
   int l,h,m;
   int i;

   if (!num_conv) return KEY_NONE;

   for (l=0,h=num_conv-1;l<h;)
   {
      m=(l+h)/2;
      c=&conv[m];
      if (!(i=k-c->gdk))
         if (c->mod)
            return c->quest|s;
         else
            return c->quest;

      if (i<0) h=m-1; else l=m+1;
   }
   c=&conv[l];
   if (k==c->gdk)
      if (c->mod)
         return c->quest|s;
      else
         return c->quest;

   return KEY_NONE;
}


#define MAX_INPUTS ((KEYM_ALT|KEYM_SHIFT|KEYM_CONTROL|0xff)+1)

static char *cmd_map[MAX_INPUTS];


static void HandleInput(int input)
{
   if (input<MAX_INPUTS && cmd_map[input])
      Cmd_Run(cmd_map[input]);
}


void S_Key_Bind(const char *key,const char *cmd)
{
   int k;

   k=Key_Parse(key);
   if (!k)
   {
      S_Error("S_Key_Bind","Couldn't parse '%s', can't bind key!",key);
      return;
   }
   if (cmd_map[k])
      free(cmd_map[k]);
   cmd_map[k]=strdup(cmd);
   if (!cmd_map[k])
   {
      S_Error("S_Key_Bind","Out of memory, can't bind '%s' to '%s'!",key,cmd);
   }
}


int S_process_key,S_process_mouse;

gint S_in_key(ARG_UNUSED(GtkWidget *w),GdkEventKey *e)
{
   int i;

   if (!S_process_key) return FALSE;
   if (e->type!=GDK_KEY_PRESS) return FALSE;
   i=S_Input_ConvKey(e->keyval,S_Input_ConvState(e->state));
   if (i) HandleInput(i);
   return FALSE;
}

gint S_in_mouse(ARG_UNUSED(GtkWidget *w),GdkEventButton *e,int *v)
{
   int i;

   if (!S_process_mouse) return FALSE;
   if (e->type!=GDK_BUTTON_PRESS) return FALSE;
   if (v) S_SetCurViewport(*v);

   i=S_Input_ConvState(e->state);
   switch (e->button)
   {
   case 1: i|=KEY_BUTTON1; break;
   case 3: i|=KEY_BUTTON2; break;
   case 2: i|=KEY_BUTTON3; break;
   default: /* shouldn't happen */ i=0; break;
   }
   HandleInput(i);

   return TRUE;
}


static int conv_cmp(const void *a,const void *b)
{
   return ((const conv_t *)a)->gdk - ((const conv_t *)b)->gdk;
}

void S_InitInput(void)
{
   S_process_key=S_process_mouse=1;
   qsort(conv,num_conv,sizeof(conv_t),conv_cmp);
}

void S_DoneInput(void)
{
}

