/*
undo.h file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#ifndef UNDO_H
#define UNDO_H


/*
These must be called every time real map objects are touched. _ALL_
functions that touch map objects must call these, including
utility functions. The caller of a function should not have to worry
about wrapping it with calls to U_xxxModify().

Calling these several times with the same object will be handled
properly. Calling these with objects that aren't in the real map
lists is OK, the calls will be ignored.

Remember to call the U_xxxModify() functions _before_ actually modifying the
object.

The U_xxxAdd and U_xxxDelete functions are called automatically by the x_Link
and x_Unlink functions. Since nobody else is supposed to link or unlink
objects, nobody else should need to call them.

*/

void U_BrushAdd(struct brush_s *b);
void U_BrushModify(struct brush_s *b);
void U_BrushDelete(struct brush_s *b);

void U_EntityAdd(struct entity_s *e);
void U_EntityModify(struct entity_s *e);
void U_EntityDelete(struct entity_s *e);

void U_GroupAdd(struct group_s *g);
void U_GroupModify(struct group_s *g);
void U_GroupDelete(struct group_s *g);


void Undo_New(const char *name,int combine,int flags);
/*
 flags:
  0=normal
  1=nothing should change until next Undo_Done
  2=clear all undo information, ignore U_* calls
*/
void Undo_Done(void);
void Undo_SetClear(void);

/* Goes over the undo information and updates everything that needs updating,
e.g. CalcBrushCenter and reallocing renderer fields. Called automatically by
Undo_Done, but also needs to be called in interactive commands if changes
are to be visible. */
void Undo_Update(void);

/* Restores everything to the way it was after the last Undo_Done (i.e.
restores what's in CurUndo and frees it). Commands can use it if they change
stuff and then want to cancel the changes (e.g. user cancelled). */
void Undo_Restore(void);

/* Like Undo_Restore, but try to recover even if map data can't be
restored. */
void Undo_Abort(void);

void Undo_Clear(void);

/* Sets flags to undo_clear */
void Undo_Init(void);

#endif

