#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "edit.h"
#include "error.h"
#include "map.h"
#include "quest.h"
#include "fe/gui.h"
#include "fe/main.h"
#include "fe/message.h"
#include "fe/sys.h"


/* Basic commands */

static void C_exit(void)
{
   if (!Map_Close("Exit"))
      return;
   S_Exit();
}
static func_t F_exit=
{"basic.exit",FUNC_VOID,{fv:C_exit},NULL,NULL,0,
{"Exit Quest."}
};


static void C_map_new(void)
{
   if (!Map_Close("New map"))
      return;

   Map_New();
/*   Cmd_SetRedraw(2); TODO */
}
static func_t F_map_new=
{"basic.map_new",FUNC_VOID,{fv:C_map_new},NULL,NULL,0,
{"Start a new map."}
};

static void C_map_load(void)
{
   char name[MAX_PATH];

   if (!Map_Close("Load map"))
      return;

   if (!S_GUI_FileDialog("Load map","Load","*.map",name))
      return;

   S_NewMessage("Loading %s...",name);
   if (!Map_Load(name))
   {
      Error_Add("map_load: Unable to load '%s'.",name);
   }

/*    Cmd_SetRedraw(2); TODO */
}
static func_t F_map_load=
{"basic.map_load",FUNC_VOID,{fv:C_map_load},NULL,NULL,0,
{"Load a map."},
};

static void C_map_save_as(void)
{
   char name[MAX_PATH];

   if (!S_GUI_FileDialog("Save map","Save","*.map",name))
      return;

   if (Sys_Exists(name))
   {
		if (!S_GUI_YesNo("Save as","File exists: '%s'\n Overwrite?",name))
         return;
   }

   Map_SetName(name);
   if (!Map_Save())
      S_NewMessage("MAP saved as %s.",M.mapfilename);
}
static func_t F_map_save_as=
{"basic.map_save_as",FUNC_VOID,{fv:C_map_save_as},NULL,NULL,0,
{"Save the current map after asking for a new name."},
};

static void C_map_save(void)
{
   Map_Save();
}
static func_t F_map_save=
{"basic.map_save",FUNC_VOID,{fv:C_map_save},NULL,NULL,0,
{"Save the current map."}
};


static func_t *funcs[]=
{&F_exit,&F_map_new,&F_map_load,&F_map_save,&F_map_save_as};

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic","Fundamental commands (exit, new, load, etc.)."}};

PLUGIN_FOOT_FUNCS

