#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "config.h"
#include "brush.h"
#include "camera.h"
#include "error.h"
#include "geom.h"
#include "quest.h"
#include "status.h"
#include "fe/3d.h"
#include "fe/int.h"
#include "fe/message.h"
#include "fe/render.h"


static config_var_t
	*cv_camera_move_drag_invert_x,
	*cv_camera_move_drag_invert_y;


/* Camera commands */


static void C_camera_zoom_in(void)
{
	S_cur_vport->zoom_amt *= cv_cam_zoom_speed->f;

	if (S_cur_vport->zoom_amt>12)
		S_cur_vport->zoom_amt=12;
	if (S_cur_vport->zoom_amt<0.125/20)
		S_cur_vport->zoom_amt=0.125/20;

	S_R_SetUpdate(R_UPDATE_CURRENT);
}
static func_t F_camera_zoom_in=
{"camera.zoom_in",FUNC_VOID,{fv:C_camera_zoom_in},NULL,NULL,0,
};

static void C_camera_zoom_out(void)
{
	S_cur_vport->zoom_amt /= cv_cam_zoom_speed->f;

	if (S_cur_vport->zoom_amt>12)
		S_cur_vport->zoom_amt=12;
	if (S_cur_vport->zoom_amt<0.125/20)
		S_cur_vport->zoom_amt=0.125/20;

	S_R_SetUpdate(R_UPDATE_CURRENT);
}
static func_t F_camera_zoom_out=
{"camera.zoom_out",FUNC_VOID,{fv:C_camera_zoom_out},NULL,NULL,0,
};


static void C_camera_reset(void)
{
	InitViewports();
	S_NewMessage("Camera Re-Initialized");
	S_R_SetUpdate(R_UPDATE_ALL);
}
static func_t F_camera_reset=
{"camera.reset",FUNC_VOID,{fv:C_camera_reset},NULL,NULL,0,
};


static void C_camera_zoom_to(void)
{
	bsel_t *bs;
	esel_t *es;
	fsel_t *fs;

	brush_t *b;
	face_t *f;
   
	vec3_t min,max;
	int found;

	int i;

	found=0;

	for (bs=M.sel.bsel;bs;bs=bs->Next)
	{
      b=bs->Brush;
		for (i=0;i<b->num_verts;i++)
		{
			if (!found)
			{
				found=1;
				max=min=b->verts[i];
			}
			else
			{
				VEC_CheckMin(min,b->verts[i]);
				VEC_CheckMax(max,b->verts[i]);
			}
		}
	}
	for (fs=M.sel.fsel;fs;fs=fs->Next)
	{
		b=fs->Brush;
		f=b->faces[fs->num];
		for (i=0;i<f->num_verts;i++)
		{
			if (!found)
			{
				found=1;
				max=min=b->verts[f->verts[i]];
			}
			else
			{
				VEC_CheckMin(min,b->verts[f->verts[i]]);
				VEC_CheckMax(max,b->verts[f->verts[i]]);
			}
		}
	}
	for (es=M.sel.esel;es;es=es->Next)
	{
		if (!found)
		{
			found=1;
			max=min=es->Entity->center;
		}
		else
		{
			VEC_CheckMin(min,es->Entity->center);
			VEC_CheckMax(max,es->Entity->center);
		}
	}

	if (!found)
	{
		S_NewMessage("cam.zoom_to: Nothing selected!");
		return;
	}

	min.x=(min.x+max.x)/2;
	min.y=(min.y+max.y)/2;
	min.z=(min.z+max.z)/2;
	Camera_AutoZoom(min);
}
static func_t F_camera_zoom_to=
{"camera.zoom_to",FUNC_VOID,{fv:C_camera_zoom_to},NULL,NULL,0,
};


static void C_camera_move(qatom_t l)
{
	int i=QINT(QLIST(&l,0));
	if (i>=MOVE_MIN && i<=MOVE_MAX)
	{
		Camera_Move(S_cur_vport,i);
		S_R_SetUpdate(R_UPDATE_CURRENT);
	}
	else
		Error("Invalid direction %i",i);
}
static func_t F_camera_move=
{"camera.move",FUNC_VOID,{fva:C_camera_move},NULL,"i",0,
};


static void C_camera_move_drag(void)
{
	s_event_t e;
	int tx,ty;
	int sx,sy,ox,oy;
	int inv_x,inv_y;

	inv_x=cv_camera_move_drag_invert_x->i?1:-1;
	inv_y=cv_camera_move_drag_invert_y->i?1:-1;

	S_I_Setup(&e,S_I_MOTION|S_I_RELEASE,NULL);
	tx=ty=0;
	ox=oy=0;
	while (1)
	{
		S_I_Wait(&e);
		if (e.type==S_I_CLICK || e.type==S_I_RELEASE)
			break;
		if (e.type==S_I_MOTION)
		{
			tx+=e.dx;
			ty+=e.dy;

			sx=-tx/4-ox;
			sy=-ty/4-oy;
			ox+=sx;
			oy+=sy;

			sx*=inv_x;
			sy*=inv_y;

			while (sx>0) Camera_Move(S_cur_vport,MOVE_LEFT),sx--;
			while (sx<0) Camera_Move(S_cur_vport,MOVE_RIGHT),sx++;
			while (sy>0) Camera_Move(S_cur_vport,MOVE_UP),sy--;
			while (sy<0) Camera_Move(S_cur_vport,MOVE_DOWN),sy++;

			S_UpdateAllViewports();
		}
	}
	S_I_Done();
}
static func_t F_camera_move_drag=
{"camera.move_drag",FUNC_VOID,{fv:C_camera_move_drag},NULL,NULL,0,
};


static void C_camera_rotate(qatom_t l)
{
	int i=QINT(QLIST(&l,0));
	vec3_t v={0,0,0};
	switch (i)
	{ /* TODO: rotate_xxx constants? */
	case MOVE_UP:       v.x=-1; break;
	case MOVE_DOWN:     v.x=+1; break;
	case MOVE_LEFT:     v.z=-1; break;
	case MOVE_RIGHT:    v.z=+1; break;
	case MOVE_FORWARD:  v.y=-1; break;
	case MOVE_BACKWARD: v.y=+1; break;
	default:
		Error("Invalid direction %i",i);
	}
	Camera_Look(v);
}
static func_t F_camera_rotate=
{"camera.rotate",FUNC_VOID,{fva:C_camera_rotate},NULL,"i",0,
};


static func_t *funcs[]={&F_camera_move,&F_camera_move_drag,
&F_camera_rotate,
&F_camera_zoom_in,&F_camera_zoom_out,
&F_camera_zoom_to,&F_camera_reset};

#include "plugin.h"

PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"camera","Fundamental camera commands (moving around, rotating, zooming)."}};

static int Init(void)
{
	cv_camera_move_drag_invert_x=Cfg_DefVar(plugin.cfg,
		"camera.move_drag.invert_x",
		"0",
		"Invert x movement when dragging camera (camera.move_drag).");
	cv_camera_move_drag_invert_y=Cfg_DefVar(plugin.cfg,
		"camera.move_drag.invert_y",
		"0",
		"Invert y movement when dragging camera (camera.move_drag).");

	Func_Register(&plugin,funcs,sizeof(funcs));

	return 1;
}

PLUGIN_FOOT

