/* TODO: split into ui-specific/non-specific part */

#include <stdio.h>
#include <stdlib.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "brush.h"
#include "config.h"
#include "edsel.h"
#include "entity.h"
#include "error.h"
#include "memory.h"
#include "quest.h"
#include "status.h"
#include "trace.h"
#include "uid.h"

#include "fe/3d.h"
#include "fe/int.h"
#include "fe/message.h"
#include "fe/popup.h"
#include "fe/trans.h"


/* TODO: clean up error-handling, try to recover partial selection? probably
not worth the trouble */


typedef struct
{
	int type;
#define T_BRUSH  0
#define T_FACE   1
#define T_ENTITY 2
	float z;
	union
	{
		brush_t *b;
		struct
		{
			brush_t *b;
			int num;
		} f;
		entity_t *e;
	} x;
} sel_object_t;


static int GetObjectList_sort(const void *v1,const void *v2)
{
	const sel_object_t *s1=v1,*s2=v2;
	if (s1->z<s2->z) return -1;
	if (s1->z>s2->z) return  1;

	return s1->type - s2->type;
}


static void GetObjectList_Wire(vec3_t m,sel_object_t **l,int *num_l)
{
	int i;
	vec3_t v;
	brush_t *b;
	face_t **f;
	entity_t *e;
	sel_object_t *l2;

	*l=NULL;
	*num_l=0;

	if (cv_render_flags->i&RENDER_BRUSH_FOCAL)
	{
		for (b=M.VBrushHead;b;b=b->VNext)
		{
			if (S_R_Trans_Vert(b->center,&v) && S_R_IsNear(m,v))
			{
				l2=Q_realloc(*l,sizeof(sel_object_t)*(*num_l+1));
				ERR_MEM(l2);
				l2[*num_l].type=T_BRUSH;
				l2[*num_l].z=v.z;
				l2[*num_l].x.b=b;
				(*num_l)++;
				*l=l2;
			}
		}
	}
	if (cv_render_flags->i&RENDER_FACE_FOCAL)
	{
		for (b=M.VBrushHead;b;b=b->VNext)
		{
			for (i=0,f=b->faces;i<b->num_faces;i++,f++)
			{
				if (S_R_Trans_Vert((*f)->center,&v) && S_R_IsNear(m,v))
				{
					l2=Q_realloc(*l,sizeof(sel_object_t)*(*num_l+1));
					ERR_MEM(l2);
					l2[*num_l].type=T_FACE;
					l2[*num_l].z=v.z;
					l2[*num_l].x.f.b=b;
					l2[*num_l].x.f.num=i;
					(*num_l)++;
					*l=l2;
				}
			}
		}
	}
	if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
	{
		for (e=M.VEntityHead;e;e=e->VNext)
		{
			if (S_R_Trans_Vert(e->center,&v) && S_R_IsNear(m,v))
			{
				l2=Q_realloc(*l,sizeof(sel_object_t)*(*num_l+1));
				ERR_MEM(l2);
				l2[*num_l].type=T_ENTITY;
				l2[*num_l].z=v.z;
				l2[*num_l].x.e=e;
				(*num_l)++;
				*l=l2;
			}
		}
	}

	if (*num_l>1)
	{
		qsort(*l,*num_l,sizeof(sel_object_t),GetObjectList_sort);
	}
}


static sel_object_t *solid_list;
static int solid_num;

static void solid_select(scalar_t dist,int type,void *v,int face)
{
	sel_object_t *l2;
	int i;

	if (type==TRACE_ENTITY)
	{
		if (((entity_t *)v)->uid==M.WorldSpawn) return; /* TODO? */
		l2=Q_realloc(solid_list,sizeof(sel_object_t)*(solid_num+1));
		ERR_MEM(l2);
		l2[solid_num].type=T_ENTITY;
		l2[solid_num].z=dist;
		l2[solid_num].x.e=v;
		solid_num++;
		solid_list=l2;

		return;
	}

	if (type==TRACE_FACE)
	{
		if (cv_render_flags->i&RENDER_FACE_FOCAL)
		{
			l2=Q_realloc(solid_list,sizeof(sel_object_t)*(solid_num+1));
			ERR_MEM(l2);
			l2[solid_num].type=T_FACE;
			l2[solid_num].z=dist;
			l2[solid_num].x.f.b=v;
			l2[solid_num].x.f.num=face;
			solid_num++;
			solid_list=l2;
		}

		if (cv_render_flags->i&RENDER_BRUSH_FOCAL)
		{
			for (i=solid_num,l2=solid_list;i;i--,l2++)
			{
				if (l2->type==T_BRUSH && l2->x.b==v)
					break;
			}
			if (i)
			{
				if (dist<l2->z) l2->z=dist;
			}
			else
			{
				l2=Q_realloc(solid_list,sizeof(sel_object_t)*(solid_num+1));
				ERR_MEM(l2);
				l2[solid_num].type=T_BRUSH;
				l2[solid_num].z=dist;
				l2[solid_num].x.b=v;
				solid_num++;
				solid_list=l2;
			}
		}
	}
}

static void GetObjectList_Solid(vec3_t m,sel_object_t **l,int *num_l)
{
	int flags;
	vec3_t org,dir;

	solid_list=NULL;
	solid_num=0;

	flags=0;
	if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
		flags|=TRACE_ENTITY;
	if (cv_render_flags->i&(RENDER_FACE_FOCAL|RENDER_BRUSH_FOCAL))
		flags|=TRACE_FACE;

	if (flags && S_R_GetVector(m,&org,&dir))
	{
		Trace(org,dir,8192,flags,solid_select);
	}

	*l=solid_list;
	*num_l=solid_num;
	if (*num_l>1)
	{
		qsort(*l,*num_l,sizeof(sel_object_t),GetObjectList_sort);
	}
}


static void GetObjectList(vec3_t m,sel_object_t **l,int *num_l)
{
	if (S_R_PreferSolid())
		GetObjectList_Solid(m,l,num_l);
	else
		GetObjectList_Wire(m,l,num_l);
}


static void SelectBrush(brush_t *b)
{
	brush_t *br;
	
	if (FindSelBrush(b))
		return;
	AddSelBrush(b);
	if (UID_Group(b->Group) && UID_Group(b->Group)->flags & GROUP_SMARTSELECT)
		for (br=M.VBrushHead; br; br=br->VNext)
			if (br->Group == b->Group)
				AddSelBrush(br);
}

static void SelectEnt(entity_t *e)
{
	entity_t *e2;
	
	if (FindSelEnt(e))
		return;
	AddSelEnt(e);
	if (UID_Group(e->Group) && UID_Group(e->Group)->flags & GROUP_SMARTSELECT)
		for (e2=M.VEntityHead; e2; e2=e2->VNext)
			if (e2->Group == e->Group)
				AddSelEnt(e2);
}


static void C_edit_select_control(void)
{
	sel_object_t *o,*p;
	int num_o;
	vec3_t m;
	int i;
	
	if (!S_R_GetMousePos(&m))
		return;

	GetObjectList(m,&o,&num_o);

	if (num_o)
	{ /* TODO: wait a while before bringing up the popup, if user releases
	     button before that, select first in list (how?) */
		int sel;
		
		if (num_o==1)
		{
			sel=0;
		}
		else
		{
			S_Popup_Init();
			for (i=0,p=o;i<num_o;i++,p++)
			{
				switch (p->type)
				{
				case T_BRUSH:
					S_Popup_AddStr("B(%g %g %g) %s",
						p->x.b->center.x,
						p->x.b->center.y,
						p->x.b->center.z,
						p->x.b->tex.name);
					break;
				case T_FACE:
					{
					face_t *f=p->x.f.b->faces[p->x.f.num];
					S_Popup_AddStr("F(%g %g %g) %s",
						f->center.x,
						f->center.y,
						f->center.z,
						f->tex.name);
					}
					break;
				case T_ENTITY:
					S_Popup_AddStr("E(%g %g %g) %s",
						p->x.e->center.x,
						p->x.e->center.y,
						p->x.e->center.z,
						GetKeyValue(p->x.e,"classname"));
					break;
				}
			}
			sel=S_Popup_Display();
			S_Popup_Free();
		}

		if (sel>=0)
		{
			switch (o[sel].type)
			{
			case T_BRUSH:
				if (FindSelBrush(o[sel].x.b))
					RemoveSelBrush(FindSelBrush(o[sel].x.b));
				else
					SelectBrush(o[sel].x.b);
				break;
			case T_FACE:
				if (FindSelFace(o[sel].x.f.b,o[sel].x.f.num))
					RemoveSelFace(FindSelFace(o[sel].x.f.b,o[sel].x.f.num));
				else
					AddSelFace(o[sel].x.f.b,o[sel].x.f.num);
				break;
			case T_ENTITY:
				if (FindSelEnt(o[sel].x.e))
					RemoveSelEnt(FindSelEnt(o[sel].x.e));
				else
					SelectEnt(o[sel].x.e);
				break;
			}
		}
		
		Q_free(o);
		return;
	}

	{
		vec3_t m2;
		vec3_t v;
		brush_t *b;
		face_t **f;
		entity_t *e;

		num_o=0;
		if (!S_I_SelectionWindow(&m,&m2))
			return;

		if (cv_render_flags->i&RENDER_BRUSH_FOCAL)
			for (b=M.VBrushHead;b;b=b->VNext)
				if (S_R_Trans_Vert(b->center,&v) && S_R_IsInBox(v,m,m2))
				{
					num_o++;
					SelectBrush(b);
				}
				
		if (cv_render_flags->i&RENDER_FACE_FOCAL)
			for (b=M.VBrushHead;b;b=b->VNext)
				for (i=0,f=b->faces;i<b->num_faces;i++,f++)
					if (S_R_Trans_Vert((*f)->center,&v) && S_R_IsInBox(v,m,m2))
					{
						num_o++;
						AddSelFace(b,i);
					}
					
		if (cv_render_flags->i&RENDER_ENTITY_FOCAL)
			for (e=M.VEntityHead;e;e=e->VNext)
			{
				if (S_R_Trans_Vert(e->center,&v) && S_R_IsInBox(v,m,m2))
				{
					num_o++;
					SelectEnt(e);
				}
			}
	}

	if (!num_o)
	{
	/* Can't find anything, so clear selection */
		ClearSelBrushes();
		ClearSelFaces();
		ClearSelEnts();
	}
}
static func_t F_edit_select_control=
{"edit.select.control",FUNC_VOID,{fv:C_edit_select_control},0,0,0,
{"Selects brushes, faces, and entities.",
"This command is used to select brushes, faces, and entities. Click on the "
"focal point of one of these objects to select it. If the object is already "
"selected, it will be deselected. Clicking where there is no focal point "
"will deselect everything."
"\n\n"
"If several focal points "
"are close to the cursor, the one closest to the camera (i.e. in front of "
"the others) will be selected. If you hold down the mouse button instead of "
"clicking, a popup window will appear that lets you select which object under "
"the cursor you wish to select. The list in the popup is sorted by distance "
"from the camera."
"\n\n"
"If you click where there is no focal point and drag, you'll get a selection "
"box. When you release the mouse button, everything inside this box will "
"be selected."}
};



static void C_edit_select_shift(void)
{
	brush_t *b;
	bsel_t *bs;
		
	face_t *f;
	fsel_t *fs;
		
	int i;
	vec3_t v;
		
	int found;
	
	{ /* First try to find a vertex where the user clicked */
		vec3_t m;

		vsel_t best;
		double best_z;
		
		if (!S_R_GetMousePos(&m))
			return;

		/* This will catch some vertices several times, but that doesn't hurt
		anything since we only select one at the end, anyway. */
		found=0;
		best_z=0;
		if (cv_render_flags->i&RENDER_BRUSH_VERT)
			for (bs=M.sel.bsel;bs;bs=bs->Next)
			{
				b=bs->Brush;
				for (i=0;i<b->num_verts;i++)
				{
					if (S_R_Trans_Vert(b->verts[i],&v) && S_R_IsNear(v,m))
					{
						if (!found)
						{
							found=1;
							best.Brush=b;
							best.num=i;
							best_z=v.z;
						}
						else
						if (v.z<best_z)
						{
							best.Brush=b;
							best.num=i;
							best_z=v.z;
						}
					}
				}
			}

		if (cv_render_flags->i&RENDER_FACE_VERT)
			for (fs=M.sel.fsel;fs;fs=fs->Next)
			{
				b=fs->Brush;
				f=b->faces[fs->num];
				for (i=0;i<f->num_verts;i++)
				{
					if (S_R_Trans_Vert(b->verts[f->verts[i]],&v) && S_R_IsNear(v,m))
					{
						if (!found)
						{
							found=1;
							best.Brush=b;
							best.num=f->verts[i];
							best_z=v.z;
						}
						else
						if (v.z<best_z)
						{
							best.Brush=b;
							best.num=f->verts[i];
							best_z=v.z;
						}
					}
				}
			}

		if (found)
		{
			if (FindSelVert(best.Brush,best.num))
				RemoveSelVert(FindSelVert(best.Brush,best.num));
			else
				AddSelVert(best.Brush,best.num);
			return;
		}
	}

	{ /* Now try a selection window */
		vec3_t m1,m2;
		
		if (!S_I_SelectionWindow(&m1,&m2))
			return;

		found=0;
		if (cv_render_flags->i&RENDER_BRUSH_VERT)
			for (bs=M.sel.bsel;bs;bs=bs->Next)
			{
				b=bs->Brush;
				for (i=0;i<b->num_verts;i++)
					if (S_R_Trans_Vert(b->verts[i],&v) && S_R_IsInBox(v,m1,m2))
					{
						AddSelVert(b,i);
						found=1;
					}
			}

		if (cv_render_flags->i&RENDER_FACE_VERT)
			for (fs=M.sel.fsel;fs;fs=fs->Next)
			{
				b=fs->Brush;
				f=b->faces[fs->num];
				for (i=0;i<f->num_verts;i++)
					if (S_R_Trans_Vert(b->verts[f->verts[i]],&v) && S_R_IsInBox(v,m1,m2))
					{
						AddSelVert(b,f->verts[i]);
						found=1;
					}
			}
		if (!found)
			ClearSelVerts();
	}
}
static func_t F_edit_select_shift=
{"edit.select.shift",FUNC_VOID,{fv:C_edit_select_shift},0,0,0,
{"Selects vertices.",
"This command selects vertices of selected brushes and faces. Click on a "
"vertex to select it. If it is already selected, it will be deselected. "
"Click and drag to get a selection box. Release the button to select "
"everything inside the box."
"\n\n"
"To deselect all vertices, click somewhere where there is no vertex "
"(or use the edit.select.clear command)."}
};


static void C_edit_select_clear(void)
{
	ClearSelAll();
}
static func_t F_edit_select_clear=
{"edit.select.clear",FUNC_VOID,{fv:C_edit_select_clear},0,0,0,
{"Deselects everything."}};


static void C_edit_select_all(void)
{
	bsel_t *bs;
	brush_t *b;

	fsel_t *fs;
	face_t *f;

	int i;

	if (cv_render_flags->i&RENDER_BRUSH_VERT)
	{
		for (bs=M.sel.bsel;bs;bs=bs->Next)
		{
			b=bs->Brush;
			for (i=0;i<b->num_verts;i++)
				AddSelVert(b,i);
		}
	}
	if (cv_render_flags->i&RENDER_FACE_VERT)
	{
		for (fs=M.sel.fsel;fs;fs=fs->Next)
		{
			b=fs->Brush;
			f=b->faces[fs->num];
			for (i=0;i<f->num_verts;i++)
				AddSelVert(b,f->verts[i]);
		}
	}
}
static func_t F_edit_select_all=
{"edit.select.all",FUNC_VOID,{fv:C_edit_select_all},0,0,0,
{"Selects all vertices on selected brushes and faces."}};


static const char help_face[]=
"If one (and only one) face is selected, these commands change the selection "
"to the next/previous face on the same brush.";
static void C_edit_select_face_next(void)
{
	brush_t *b;
	int num;
	
	if (!M.sel.fsel || M.sel.fsel->Next)
	{
		S_NewMessage("edit.select.face_next: Exactly one face must be selected!");
		return;
	}
	b=M.sel.fsel->Brush;
	num=M.sel.fsel->num;
	RemoveSelFace(M.sel.fsel);
	num+=1;
	if (num==b->num_faces) num=0;
	AddSelFace(b,num);
}
static func_t F_edit_select_face_next=
{"edit.select.face_next",FUNC_VOID,{fv:C_edit_select_face_next},0,0,0,
{"Selects the next face on the brush.",help_face}};

static void C_edit_select_face_prev(void)
{
	brush_t *b;
	int num;
	
	if (!M.sel.fsel || M.sel.fsel->Next)
	{
		S_NewMessage("edit.select.face_prev: Exactly one face must be selected!");
		return;
	}
	b=M.sel.fsel->Brush;
	num=M.sel.fsel->num;
	RemoveSelFace(M.sel.fsel);
	num-=1;
	if (num==-1) num=b->num_faces-1;
	AddSelFace(b,num);
}
static func_t F_edit_select_face_prev=
{"edit.select.face_prev",FUNC_VOID,{fv:C_edit_select_face_prev},0,0,0,
{"Selects the previous face on the brush.",help_face}};


static func_t *funcs[]=
{&F_edit_select_control,&F_edit_select_shift,
&F_edit_select_all,&F_edit_select_clear,
&F_edit_select_face_next,&F_edit_select_face_prev};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"edit.select","Basic commands to select stuff in viewports."}};

PLUGIN_FOOT_FUNCS

