#include <stdio.h>

#include "defines.h"
#include "types.h"

#include "qfunc.h"

#include "brush.h"
#include "camera.h"
#include "edsel.h"
#include "entclass.h"
#include "entity.h"
#include "error.h"
#include "geom.h"
#include "map.h"
#include "memory.h"
#include "quest.h"
#include "uid.h"
#include "undo.h"

#include "fe/3d.h"
#include "fe/gui.h"
#include "fe/message.h"


static void C_entity_add_brush(void)
{
	bsel_t *b;
	entity_t *e;

	if (!M.sel.num_bselected)
	{
		S_NewMessage("entity.add_brush: At least one brush must be selected.");
		return;
	}

	if (M.sel.num_eselected!=1)
	{
		S_NewMessage("entity.add_brush: Exactly one entity must be selected.");
		return;
	}

	e=M.sel.esel->Entity;
	for (b=M.sel.bsel;b;b=b->Next)
	{
		U_BrushModify(b->Brush);
		b->Brush->Entity = e->uid;
	}
}
static func_t F_entity_add_brush=
{"entity.add_brush",FUNC_VOID,{fv:C_entity_add_brush},NULL,NULL,0,
{"Add the selected brushes to the selected entity."}};


static void C_entity_create(void)
{
	char buf[256];
	entity_t *e;
	vec3_t d;

	if (!S_GUI_EntityPicker(buf,ECLASS_POINT))
		return;

	e=E_New();
	ERR_MEM(e);
	SetKeyValue(e,"classname",buf);

	Move(S_cur_vport,MOVE_FORWARD,&d,100); /* TODO: edit.project_object */
	VEC_Add(d,d,S_cur_vport->camera_pos);
	SnapVecToGrid(&d);
	sprintf(buf,"%i %i %i",(int)d.x,(int)d.y,(int)d.z);
	SetKeyValue(e,"origin",buf);

	SetEntityDefaults(e);
	E_Link(e);

	ClearSelAll();
	AddSelEnt(e);
}
static func_t F_entity_create=
{"entity.create",FUNC_VOID,{fv:C_entity_create},NULL,NULL,0,
{"Create an entity."}};


static void C_entity_create_model(void)
{
	char classname[256];
	entity_t *e;
	bsel_t *bs;

	if (M.sel.num_bselected == 0)
	{
		S_NewMessage("entity.create_model: At least one brush must be selected.");
		return;
	}
	if (!S_GUI_EntityPicker(classname,ECLASS_MODEL))
		return;

	e=E_New();
	ERR_MEM(e);

	SetKeyValue(e, "classname", classname);
	SetEntityDefaults(e);

	E_Link(e);

	/* Assign the brushes to the entity */
	for (bs=M.sel.bsel; bs; bs=bs->Next)
	{
		U_BrushModify(bs->Brush);
		bs->Brush->Entity = e->uid;
	}

	ClearSelBrushes();
}
static func_t F_entity_create_model=
{"entity.create_model",FUNC_VOID,{fv:C_entity_create_model},NULL,NULL,0,
{"Create a model."}};


static void C_entity_edit(void)
{
	entity_t **ents;
	int nents;
	esel_t *e;
	bsel_t *bs;
	map_uid_t eu;
	int i;

	if (!M.sel.esel && !M.sel.bsel)
	{
		S_NewMessage("entity.edit: Nothing selected!");
		return;
	}

	nents=0;
	ents=NULL;
	for (e=M.sel.esel;e;e=e->Next)
	{
		nents++;
		ents=Q_realloc(ents,sizeof(entity_t *)*nents);
		ERR_MEM(ents);
		ents[nents-1]=e->Entity;
	}
	for (bs=M.sel.bsel;bs;bs=bs->Next)
	{
		eu=bs->Brush->Entity;
		for (i=0;i<nents;i++)
			if (eu==ents[i]->uid)
				break;
		if (i!=nents) continue;
		nents++;
		ents=Q_realloc(ents,sizeof(entity_t *)*nents);
		ERR_MEM(ents);
		ents[nents-1]=UID_Entity(eu);
	}

	E_Edit(nents,ents);

	Q_free(ents);
}
static func_t F_entity_edit=
{"entity.edit",FUNC_VOID,{fv:C_entity_edit},NULL,NULL,0,
{"Edit the selected entities and the entities of the selected brushes."}};


static func_t *funcs[]={&F_entity_add_brush,&F_entity_create,
	&F_entity_create_model,&F_entity_edit};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"entity.basic","Basic entity commands."}};

PLUGIN_FOOT_FUNCS

