/*
cmd/fsplit.c file of the Quest Source Code

Copyright 1997, 1998 Alexander Malmberg
Copyright 1996, Trey Harrison and Chris Carollo

This program is distributed under the GNU General Public License.
See legal.txt for more information.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "defines.h"
#include "types.h"

#include "cmds.h"

#include "brush.h"
#include "edsel.h"
#include "game.h"
#include "memory.h"
#include "quest.h"
#include "s_error.h"
#include "s_int.h"
#include "undo.h"


/*
TODO: update with new brush/face	structures
*/

static void SplitFace(int v1,int v2)
{
   brush_t *b;
   int e1;
   int i,j;
   face_t *f1,*f2;
   int nv1;
   int *verts1;
   int nv2;
   int *verts2;

   b=M.sel.fsel->Brush;

   if (v2<v1)
   {
      int i;
      i=v2;
      v2=v1;
      v1=i;
   }

   for (i=0;i<b->num_edges;i++)
   {
      if ((b->edges[i].startvertex==v1 && b->edges[i].endvertex==v2)||
          (b->edges[i].startvertex==v2 && b->edges[i].endvertex==v1))
      {
         S_Error("SplitFace","Vertices on same line");
         return;
      }
   }

   U_BrushModify(b);
   
   e1=b->num_edges++;

   b->edges=Q_realloc(b->edges,sizeof(edge_t)*b->num_edges);
   if (!b->edges)
      Cmd_Abort("SplitFace","Out of memory!");

   b->edges[e1].startvertex=v1;
   b->edges[e1].endvertex=v2;

   b->num_faces++;
   b->faces=Q_realloc(b->faces,sizeof(face_t)*b->num_faces);
   if (!b->faces)
      Cmd_Abort("SplitFace","Out of memory!");

   f1=&b->faces[M.sel.fsel->num];
   f2=&b->faces[b->num_faces-1];

   if (!F_CopyP(f2,f1))
      Cmd_Abort("SplitFace","Out of memory!");

   nv1=nv2=0;
   verts1=verts2=NULL;
   j=0;

   for (i=0;i<f1->num_verts;i++)
   {
      if (!j)
      {
         nv1++;
         verts1=Q_realloc(verts1,nv1*sizeof(int));
         if (!verts1)
            Cmd_Abort("SplitFace","Out of memory!");
         verts1[nv1-1]=f1->verts[i];
         if (f1->verts[i]==v1 || f1->verts[i]==v2)
         {
            nv2++;
            verts2=Q_realloc(verts2,nv2*sizeof(int));
            if (!verts2)
               Cmd_Abort("SplitFace","Out of memory!");
            verts2[nv2-1]=f1->verts[i];
            j++;
         }
      }
      else
      {
         nv2++;
         verts2=Q_realloc(verts2,nv2*sizeof(int));
         if (!verts2)
            Cmd_Abort("SplitFace","Out of memory!");
         verts2[nv2-1]=f1->verts[i];
         if (f1->verts[i]==v1 || f1->verts[i]==v2)
         {
            nv1++;
            verts1=Q_realloc(verts1,nv1*sizeof(int));
            if (!verts1)
               Cmd_Abort("SplitFace","Out of memory!");
            verts1[nv1-1]=f1->verts[i];
            j=0;
         }
      }
   }

   Q_free(f1->verts);

   f1->verts=verts1;
   f1->num_verts=nv1;
   f2->verts=verts2;
   f2->num_verts=nv2;
}


static int SplitLine(void)
{
   brush_t *b;
   face_t *f;
   int v1,v2,v3;
   int i,j,k;
   int e1,e2;

   if (M.sel.num_fselected!=1 || M.sel.num_bselected)
   {
      S_Error("SplitLine","Exactly one face must be selected!");
      return -1;
   }
   if (M.sel.num_vselected!=2)
   {
      S_Error("SplitLine","Exactly two vertices must be selected!");
      return -1;
   }

   b=M.sel.fsel->Brush;
   if (!(b->bt->flags&BRUSH_F_NORMAL))
   {
      S_Error("SplitLine","Can only split normal brushes");
      return -1;
   }

   v1=M.sel.vsel->num;
   v2=M.sel.vsel->Next->num;

   if (v2<v1)
   {
      int i;
      i=v2;
      v2=v1;
      v1=i;
   }

   for (i=0;i<b->num_edges;i++)
   {
      if ((b->edges[i].startvertex==v1 && b->edges[i].endvertex==v2)||
          (b->edges[i].startvertex==v2 && b->edges[i].endvertex==v1))
         break;
   }
   if (i==b->num_edges)
   {
      S_Error("SplitLine","Vertices not on same line");
      return -1;
   }
   e1=i;

   U_BrushModify(b);

   v3=b->num_verts++;

   b->verts=Q_realloc(b->verts,sizeof(vec3_t)*b->num_verts);
   if (!b->verts)
      Cmd_Abort("SplitLine","Out of memory!");

   b->verts[v3].x=(b->verts[v1].x+b->verts[v2].x)/2;
   b->verts[v3].y=(b->verts[v1].y+b->verts[v2].y)/2;
   b->verts[v3].z=(b->verts[v1].z+b->verts[v2].z)/2;

   e2=b->num_edges++;

   b->edges=Q_realloc(b->edges,sizeof(edge_t)*b->num_edges);
   if (!b->edges)
      Cmd_Abort("SplitLine","Out of memory!");

   if (b->edges[e1].startvertex==v1)
   {
      b->edges[e1].endvertex=v3;

      b->edges[e2].startvertex=v3;
      b->edges[e2].endvertex=v2;
   }
   else
   {
      b->edges[e1].endvertex=v3;

      b->edges[e2].startvertex=v3;
      b->edges[e2].endvertex=v1;
   }

   for (i=0;i<b->num_faces;i++)
   {
      f=&b->faces[i];
      for (j=0;j<f->num_verts-1;j++)
      {
         if ((f->verts[j]==v1 && f->verts[j+1]==v2) ||
             (f->verts[j]==v2 && f->verts[j+1]==v1))
         {
            f->num_verts++;
            f->verts=Q_realloc(f->verts,f->num_verts*sizeof(int));
            if (!f->verts)
               Cmd_Abort("SplitLine","Out of memory!");

            for (k=f->num_verts-1;k>j+1;k--)
               f->verts[k]=f->verts[k-1];
            f->verts[j+1]=v3;
         }
      }
      if ((f->verts[0]==v1 && f->verts[f->num_verts-1]==v2) ||
          (f->verts[0]==v2 && f->verts[f->num_verts-1]==v1))
      {
         f->num_verts++;
         f->verts=Q_realloc(f->verts,f->num_verts*sizeof(int));
         if (!f->verts)
            Cmd_Abort("SplitLine","Out of memory!");
         f->verts[f->num_verts-1]=v3;
      }
   }

   Cmd_SetRedraw(2);

   return v3;
}


static void C_split_edge_and_face(void)
{
   int v1,v2;

   if (M.sel.num_fselected!=1 || M.sel.num_bselected)
   {
      S_Error("face.split_edge_and_face",
         "Exactly one face must be selected");
      return;
   }

   v1=SplitLine();
   if (v1==-1)
      return;

   ClearSelVerts();
   S_I_GetSelection("Select the other edge.");

   v2=SplitLine();
   if (v2==-1)
      return;

   SplitFace(v1,v2);

   ClearSelVerts();
   Cmd_SetRedraw(2);
}


static void C_split(void)
{
   brush_t *b;
   int v1,v2;

   if (M.sel.num_fselected!=1 || M.sel.num_bselected)
   {
      S_Error("face.split","Exactly one face must be selected!");
      return;
   }
   if (M.sel.num_vselected!=2)
   {
      S_Error("face.split","Exactly two vertices must be selected!");
      return;
   }

   b=M.sel.fsel->Brush;
   if (!(b->bt->flags&BRUSH_F_NORMAL))
   {
      S_Error("face.split","Can only split normal brushes");
      return;
   }

   v1=M.sel.vsel->num;
   v2=M.sel.vsel->Next->num;

   SplitFace(v1,v2);

   ClearSelVerts();

   Cmd_SetRedraw(2);
}

static void C_split_edge(void)
{
   SplitLine();
}


static cmd_reg_t cmds[]={
{"face.split"              ,C_split              ,{"Brush/Face/Split/Face","control+f"},{},"Split the selected face."},
{"face.split_edge"         ,C_split_edge         ,{"Brush/Face/Split/Edge","control+l"},{},"Split the selected edge (two vertices)."},
{"face.split_edge_and_face",C_split_edge_and_face,{"Brush/Face/Split/Face and edge","alt+s"},{},"Split two edges and a face."},
};

#include "plugin.h"
PLUGIN_HEAD

static plugin_t plugin=
{{Init},{"basic/facesplit","Commands to split faces and edges."}};

PLUGIN_FOOT_CMDS

