#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include <gdk/gdk.h>
#include <gtk/gtk.h>
#include <gtkgl/gtkglarea.h>

/* these are redefined in defines.h */
#undef TRUE
#undef FALSE

#include "quest/defines.h"
#include "quest/types.h"

#include "../../gl/gl_rend.h"

#include "g_3d.h"
#include "g_3d_all.h"
#include "../input.h"

#include "quest/fe/trans.h"

#include "quest/fe/int.h"

#include "quest/error.h"
#include "quest/ql.h"
#include "quest/quest.h"
#include "quest/fe/texture.h"

#include "../../gl/gl_texture.h"


static g_viewport_t **vports;
static int num_vports;


viewport_t *S_cur_vport;
g_viewport_t *S_cur_gvport;
gl_viewport_t *S_GL_vport;
static int cur_vport;


static void init(g_viewport_t *v)
{
   if (!gtk_gl_area_make_current(v->gla))
      return;
   S_GL_Init(&v->v);
}

static void resize(g_viewport_t *v)
{
   if (!gtk_gl_area_make_current(v->gla))
      return;
   S_GL_Resize(&v->v);
}

static void draw(g_viewport_t *v)
{
   if (!v->v.vp.active) return;

   if (!gtk_gl_area_make_current(v->gla))
      return;
   S_GL_Render(&v->v);
   gtk_gl_area_swap_buffers(v->gla);
}


void S_SetCurViewport(int v)
{
   if (cur_vport==v) return;
   cur_vport=v;
   if (!vports[v]) ABORT("S_SetCurViewport: Tried to use stale viewport!"); /* TODO? */
   S_cur_gvport=vports[cur_vport];
   S_cur_vport=&S_cur_gvport->v.vp;
   S_GL_vport=&S_cur_gvport->v;
}


void S_UpdateCurViewport(void)
{
   draw(S_cur_gvport);
}

void S_UpdateAllViewports(void)
{
   int i;
   g_viewport_t **v;

   for (v=vports,i=num_vports;i;i--,v++)
      if (*v)
         draw(*v);
}


void S_R_ForEachViewport(void (*func)(struct viewport_s *))
{
   int i;

   for (i=0;i<num_vports;i++)
      if (vports[i])
         func(&vports[i]->v.vp);
}


static gint vp_configure(GtkWidget *w,ARG_UNUSED(GdkEventConfigure *e),g_viewport_t *vp)
{
   vp->v.sx=w->allocation.width;
   vp->v.sy=w->allocation.height;

   resize(vp);
   draw(vp);
      
   return TRUE;
}

static gint vp_expose(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEventExpose *e),g_viewport_t *vp)
{
   draw(vp);
   return FALSE;
}

static gint vp_focus(ARG_UNUSED(GtkWidget *w),GdkEventFocus *e,g_viewport_t *v)
{
   if (e->in && v->num!=cur_vport)
      S_SetCurViewport(v->num);
   return FALSE;
}


static gint vp_delete(ARG_UNUSED(GtkWidget *w),ARG_UNUSED(GdkEvent *e),g_viewport_t *v)
{
   int i;
   
   if (num_vports==1)
   { /* Can't delete last viewport, generate an 'exit' command instead. */
/* TODO      Cmd_Run("exit"); */
      return TRUE;
   }

   for (i=v->num;i<num_vports-1;i++)
   {
      vports[i]=vports[i+1];
      vports[i]->num=i;
   }
   num_vports--;
   free(v);

   cur_vport=-1;
   S_SetCurViewport(0);
   
   return FALSE;
}


/*
First created viewport needs to be handled specially as the
renderer isn't setup yet, and it can't be until it can create
the necessary contexts and stuff.

Returns number of new viewport or -1 on error.
*/
static int NewViewport(int px,int py,int x,int y,int first)
{
   g_viewport_t *vp;
   int num;

   {
      g_viewport_t **v;
      v=realloc(vports,sizeof(g_viewport_t *)*(num_vports+1));
      if (!v)
      {
         if (first)
            ABORT("FirstViewport: Unable to allocate memory for new viewport!");
         else
         { /* TODO */
            Warning("NewViewport: Unable to allocate memory for new viewport!");
            return -1;
         }
      }
      vports=v;
      num=num_vports;
   }
   vp=malloc(sizeof(g_viewport_t));
   if (!vp)
   {
      if (first)
         ABORT("FirstViewport: Unable to allocate memory for new viewport!");
      else
      {
         Warning("NewViewport: Unable to allocate memory for new viewport!");
         return -1;
      }
   }
   vports[num]=vp;

   memset(vp,0,sizeof(g_viewport_t));

   vp->num=num;

   vp->v.sx=x;
   vp->v.sy=y;

   {
static int attribs[]=
{
GDK_GL_RGBA,
GDK_GL_DOUBLEBUFFER,
GDK_GL_RED_SIZE,1,
GDK_GL_GREEN_SIZE,1,
GDK_GL_BLUE_SIZE,1,
GDK_GL_DEPTH_SIZE,1,
GDK_GL_NONE
};
      if (first)
         vp->gla=GTK_GL_AREA(gtk_gl_area_new(attribs));
      else
         vp->gla=GTK_GL_AREA(gtk_gl_area_share_new(attribs,vports[0]->gla));
      if (!vp->gla)
      {
         if (first)
            ABORT("FirstViewport: Couldn't create gl context!");
         else
         {
            Warning("NewViewport: Couldn't create gl context!");
            return -1;
         }
      }
   }

   vp->win=GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
   gtk_window_set_wmclass(vp->win,"quest","viewport");
   gtk_window_set_title(vp->win,"Viewport");
   gtk_window_set_policy(vp->win,1,1,0);
   gtk_window_set_transient_for(vp->win,S_menu_window);

   gtk_signal_connect(GTK_OBJECT(vp->win),"delete_event",GTK_SIGNAL_FUNC(vp_delete),vp);

   gtk_signal_connect(GTK_OBJECT(vp->win),"focus_in_event",GTK_SIGNAL_FUNC(vp_focus),vp);

   gtk_widget_set_events(GTK_WIDGET(vp->win),
      GDK_BUTTON_PRESS_MASK|GDK_BUTTON_RELEASE_MASK|
      GDK_KEY_PRESS_MASK|GDK_KEY_RELEASE_MASK);
   gtk_signal_connect(GTK_OBJECT(vp->win),"key_press_event",GTK_SIGNAL_FUNC(S_in_key),NULL);
   gtk_signal_connect(GTK_OBJECT(vp->win),"key_release_event",GTK_SIGNAL_FUNC(S_in_key),NULL);
   gtk_signal_connect(GTK_OBJECT(vp->win),"button_press_event",GTK_SIGNAL_FUNC(S_in_mouse),&vp->num);
   gtk_signal_connect(GTK_OBJECT(vp->win),"button_release_event",GTK_SIGNAL_FUNC(S_in_mouse),&vp->num);

   gtk_widget_set_events(GTK_WIDGET(vp->gla),GDK_EXPOSURE_MASK);
   gtk_widget_set_usize(GTK_WIDGET(vp->gla),x,y);
   gtk_container_add(GTK_CONTAINER(vp->win),GTK_WIDGET(vp->gla));


   if (first)
   {
      gtk_widget_realize(GTK_WIDGET(vp->gla));
      gtk_widget_realize(GTK_WIDGET(vp->win));
      if (!gtk_gl_area_make_current(vp->gla))
      {
         ABORT("FirstViewport: Unable to activate OpenGL context!");
      }
      S_GL_InitRender();
   }


   gtk_signal_connect(GTK_OBJECT(vp->gla),"configure_event",GTK_SIGNAL_FUNC(vp_configure),vp);
   gtk_signal_connect(GTK_OBJECT(vp->gla),"expose_event",GTK_SIGNAL_FUNC(vp_expose),vp);

   gtk_widget_show(GTK_WIDGET(vp->gla));
   gtk_widget_show(GTK_WIDGET(vp->win));

   if (px!=-1 && py!=-1)
   { /* TODO: do without using 'internal' function */
      gtk_window_reposition(vp->win,px,py);
   }

   num_vports++;
   
   init(vp);

   return num;
}


#include "quest/camera.h"
#include "quest/qfunc.h"

static void C_gtk_gl_new_viewport(void)
{
	int i;
	if ((i=NewViewport(-1,-1,320,240,0)))
	{
		InitViewport(&vports[i]->v.vp);
		draw(vports[i]);
	}
}
static func_t F_gtk_gl_new_viewport=
{"gtk_gl.new_viewport",FUNC_VOID,{fv:C_gtk_gl_new_viewport},NULL,NULL,0,
{"Create a new viewport."}};


#include "quest/config.h"

static config_var_t *cv_gtk_win_viewports;

void S_InitViewports(void)
{
	const char *str;
	const char *c;
	int first;
	int x,y,sx,sy;

	cv_gtk_win_viewports=Cfg_DefVar(config_fe,"gtk_gl.win.viewports","(-1 -1 320 240)",NULL);

	cur_vport=-1;
	first=1;
	str=cv_gtk_win_viewports->s;
	for (c=strchr(str,'(');c;c=strchr(c,'('))
	{
		c++;
		if (sscanf(c,"%i %i %i %i",&x,&y,&sx,&sy)!=4) break;
		if (first)
		{
			S_SetCurViewport(NewViewport(x,y,sx,sy,1));
			first=0;
		}
		else
			NewViewport(x,y,sx,sy,0);
	}
	if (first)
	{
		S_SetCurViewport(NewViewport(-1,-1,320,240,1));
	}
   
	Func_RegisterOne(NULL,&F_gtk_gl_new_viewport);
}


/* TODO: window manager weirdness causes windows to 'drift'
highly annoying */
void S_DoneViewports(void)
{
	int i;
	char buf[256];
	int x,y,sx,sy;
   
	S_GL_DoneRender();
	S_GL_Texture_Done();
	buf[0]=0;
	for (i=0;i<num_vports;i++)
	{
		if (!vports[i]) continue;

		gdk_window_get_root_origin(GTK_WIDGET(vports[i]->win)->window,&x,&y);
		gdk_window_get_size(GTK_WIDGET(vports[i]->win)->window,&sx,&sy);
		sprintf(&buf[strlen(buf)]," (%i %i %i %i)",x,y,sx,sy);
      
		gtk_widget_destroy(GTK_WIDGET(vports[i]->gla));
		gtk_widget_destroy(GTK_WIDGET(vports[i]->win));
	}
	Cfg_SetStr(cv_gtk_win_viewports,buf);
}


int S_R_GetMousePos(vec3_t *spos)
{
   int x,y;

   gdk_window_get_pointer(GTK_WIDGET(S_cur_gvport->win)->window,&x,&y,NULL);
   if (x<0 || y<0 || x>=S_cur_gvport->v.sx || y>=S_cur_gvport->v.sy)
      return 0;
   spos->x=  (x/(double)S_cur_gvport->v.sx)*2-1;
   spos->y=-((y/(double)S_cur_gvport->v.sy)*2-1);
   spos->z=0;
   return 1;
}

