/* Included from g_3dr.c */


#define OutCode3D(x,y,z,code)            \
{                                        \
	(code)=0;                             \
	if(((x)-0.0001)>(z))  (code) |= 0x01; \
	if(((x)+0.0001)<-(z)) (code) |= 0x02; \
	if(((y)-0.0001)>(z))  (code) |= 0x04; \
	if(((y)+0.0001)<-(z)) (code) |= 0x08; \
	if(((z)-0.0001)>1)    (code) |= 0x10; \
	if(((z)+0.0001)<0)    (code) |= 0x20; \
}


static void Clip2D(int x0,int y0,int x1,int y1,int *x2,int *y2,int code)
{
   float dx, dy;

   dx = x1 - x0;
   dy = y1 - y0;

   if (code & 0x0001)
   {
      *x2 = 0;
      *y2 = (float)y0 + (float)(*x2 - x0) * dy / dx;
   }
   else
   if (code & 0x0002)
   {
      *x2 = cur_v->sx;
      *y2 = (float)y0 + (float)(*x2 - x0) * dy / dx;
   }
   else
   if (code & 0x0004)
   {
      *y2 = 0;
      *x2 = (float)x0 + (float)(*y2 - y0) * dx / dy;
   }
   else
   if (code & 0x0008)
   {
      *y2 = cur_v->sy;
      *x2 = (float)x0 + (float)(*y2 - y0) * dx / dy;
   }
}

static void Clip3D(float  x0,float  y0,float  z0,float  x1,float  y1,float  z1,float *x2,float *y2,float *z2,float *t,int code)
{
   float dx, dy, dz;
   float sc;

   dx = x1 - x0;
   dy = y1 - y0;
   dz = z1 - z0;

   if (code & 0x00000001)
      sc = (-x0 + z0) / (dx - dz);
   else
   if (code & 0x00000002)
      sc = (x0 + z0) / (-dx - dz);
   else
   if (code & 0x00000004)
      sc = (-y0 + z0) / (dy - dz);
   else
   if (code & 0x00000008)
      sc = (y0 + z0) / (-dy - dz);
   else
   if (code & 0x00000010)
      sc = (1 - z0) / (dz);
   else
//   if (code & 0x00000020)
      sc = (z0) / (-dz);

   *x2 = x0 + (sc * dx);
   *y2 = y0 + (sc * dy);
   *z2 = z0 + (sc * dz);
   *t = sc;
}


static int OutCode2D(int x,int y)
{
   int code = 0;

   if (x < 0)         code |= 0x0001;
   if (x > cur_v->sx) code |= 0x0002;
   if (y < 0)         code |= 0x0004;
   if (y > cur_v->sy) code |= 0x0008;

   return code;
}


static int OutsideFrustrum(vec3_t from, vec3_t to)
{
   if ((from.z <= 0)      && (to.z <= 0))    return TRUE;
   if ((from.x >  from.z) && (to.x >  to.z)) return TRUE;
   if ((from.x < -from.z) && (to.x < -to.z)) return TRUE;
   if ((from.y >  from.z) && (to.y >  to.z)) return TRUE;
   if ((from.y < -from.z) && (to.y < -to.z)) return TRUE;
   return FALSE;
}


static void TransformLineToScreen3D(
   float x0,float y0,float z0,
   float x1,float y1,float z1,
   int *px0,int *py0,int *px1,int *py1)
{
   if ((z0 < 0.01) && (x0 < 0.01) && (y0 < 0.01))
      x0 = y0 = 0;
   else
   {
      x0 /= z0;
      y0 /= z0;
   }
   if ((z1 < 0.01) && (x1 < 0.01) && (y1 < 0.01))
      x1 = y1 = 0;
   else
   {
      x1 /= z1;
      y1 /= z1;
   }

   x0 = dx * (x0 + 1) + 1;
   y0 = dy * (y0 + 1) + 1;
   x1 = dx * (x1 + 1) + 1;
   y1 = dy * (y1 + 1) + 1;

   *px0=x0;
   *px1=x1;
   *py0=y0;
   *py1=y1;
}

static int ClipLine3D(
   float x0,float y0,float z0,
   float x1,float y1,float z1,
   int *px0,int *py0,int *px1,int *py1)
{
   float dx, dy, dz, t=0;
// float s;
   int c0;
   int c1;
// int clip0, clip1;

   y0 = -y0;
   y1 = -y1;

restart_clip3d:

   OutCode3D(x0, y0, z0, c0);
   OutCode3D(x1, y1, z1, c1);

   if (!cur_v->depth_clip)
   {
      c0 &= 0xffffffef;
      c1 &= 0xffffffef;
   }

   if ((c0 | c1) == 0)
   {
      TransformLineToScreen3D(
         x0, y0, z0, x1, y1, z1,
         px0,py0,px1,py1);
      return 1;
   }
   else
   if ((c0 & c1) != 0)
   {
      return 0;
   }
   else
   {
      if (c0 != 0)
      {
         dx = x1 - x0;
         dy = y1 - y0;
         dz = z1 - z0;
         if (c0 & 0x00000001)
            t = (-x0 + z0) / (dx - dz);
         else
         if (c0 & 0x00000002)
            t = (x0 + z0) / (-dx - dz);
         else
         if (c0 & 0x00000004)
            t = (-y0 + z0) / (dy - dz);
         else
         if (c0 & 0x00000008)
            t = (y0 + z0) / (-dy - dz);
         else
         if (c0 & 0x00000010)
            t = (1 - z0) / (dz);
         else
         if (c0 & 0x00000020)
            t = (z0) / (-dz);
         x0 = x0 + (t * dx);
         y0 = y0 + (t * dy);
         z0 = z0 + (t * dz);
      }
      else
      {
         dx = x1 - x0;
         dy = y1 - y0;
         dz = z1 - z0;
         if (c1 & 0x00000001)
            t = (-x0 + z0) / (dx - dz);
         else
         if (c1 & 0x00000002)
            t = (x0 + z0) / (-dx - dz);
         else
         if (c1 & 0x00000004)
            t = (-y0 + z0) / (dy - dz);
         else
         if (c1 & 0x00000008)
            t = (y0 + z0) / (-dy - dz);
         else
         if (c1 & 0x00000010)
            t = (1 - z0) / (dz);
         else
         if (c1 & 0x00000020)
            t = (z0) / (-dz);
         x1 = x0 + (t * dx);
         y1 = y0 + (t * dy);
         z1 = z0 + (t * dz);
      }
      goto restart_clip3d;
   }
}

static void ClipDrawLine3D(float x0,float y0,float z0,float x1,float y1,float z1)
{
   int dx0,dy0,dx1,dy1;

   if (ClipLine3D(x0,y0,z0,x1,y1,z1,&dx0,&dy0,&dx1,&dy1))
      DrawLine(dx0,dy0,dx1,dy1);
}

