#ifndef CSG_H
#define CSG_H

#include "brush.h" /* TODO? */


#define MAX_FACE_POINTS 64   /* TODO */

typedef struct
{
	char    misc;
	vec3_t  planepts[3];
   plane_t p;
	int     num_pts;
	vec3_t  pts[MAX_FACE_POINTS];  /* points on face / plane (the polygon pts) */

   texdef_t tex;
   brush_normal_align_t ba;
} cface_t;

typedef struct
{
	int num_faces;
	cface_t *faces;
} cbrush_t;


/* Fills in p and pts, creates new copy of tex.g */
cface_t *CSG_FaceToCFace(cface_t *dst,const struct brush_s *b,int num);
cbrush_t *CSG_BrushToCBrush(cbrush_t *dst,const struct brush_s *b);

void CSG_FreeFace(cface_t *f);
void CSG_FreeBrush(cbrush_t *b);


/* Calculates p from planepts */
int CSG_CalcPlane(cface_t *f);
void CSG_CalcPlanes(cbrush_t *b);


/* Creates a huge box on the face's plane */
void CSG_MakeBoxOnPlane(cface_t *f);

/* Fills in pts, and lots of other stuff. 1 on failure */
int CSG_MakeFaces(cbrush_t *b);

/* Builds the brush, doesn't set Entity or Group. Q_free:s b->faces. Copies
tex.g pointers, not contents. */
struct brush_s *CSG_BuildBrush(struct brush_s *dst,cbrush_t *b);

/*
Returns:
  0 - Face was clipped successfully
 -1 - Face was completely behind the plane
 -2 - Face was completely in front of the plane
 -3 - Error clipping / culling face
 -4 - Face was completely ON the plane
*/
int CSG_SplitFace(cface_t *f,plane_t p,cface_t *front,cface_t *back,int keepback);

/*
  0 - Plane intersects brush
  1 - On and in front
 -1 - On and behind
 -2 - Error (empty brush)
*/
int CSG_BrushSide(cbrush_t *b,plane_t p);

int CSG_BrushesIntersect(cbrush_t *b1,cbrush_t *b2);

/*
   0 - The Brush (bfront) was split into 2 brushes, a front and a back
  -1 - The Brush (bfront) was entirely in front of the plane, not split
  -2 - The Brush (bfront) was entirely behind the plane, not split
  -3 - Error
*/
int CSG_SplitBrushWithFace(cbrush_t *Bfront, cface_t p,cbrush_t *Bback);

#endif

