#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qatom.h"

#include "error.h"
#include "memory.h"
#include "qsymbol.h"


qatom_t qa_undef={QAT_UNDEF};
qatom_t qa_nil={QAT_SYMBOL};
qatom_t qa_T={QAT_SYMBOL}; /* .x.symbol is set properly in QL_Init */


void QA_Free(qatom_t a)
{
	if (QAT_ISLIST(a.type))
	{
		int i;
		for (i=0;i<a.type;i++)
			QA_Free(a.x.l[i]);
		Q_free(a.x.l);
		return;
	}
	switch (a.type)
	{
	case QAT_SYMBOL:
	case QAT_FLOAT:
	case QAT_INT:
	case QAT_FUNC:
	case QAT_MENU:
	case QAT_UNDEF:
		break;
	case QAT_STRING:
		Q_free(a.x.string);
		break;
	default:
		ERROR("QA_free: unknown atom type %i",a.type);
	}
}

qatom_t QA_Copy(qatom_t a)
{
	qatom_t b;
	int i;
	b.type=a.type;
	if (QAT_ISLIST(a.type))
	{
		if (a.type>0)
		{
			b.x.l=Q_malloc(sizeof(qatom_t)*b.type);
			ERR_MEM(b.x.l);
		}
		else
			b.x.l=NULL;
		for (i=0;i<b.type;i++)
			b.x.l[i]=QA_Copy(a.x.l[i]);
		return b;
	}
	switch (a.type)
	{
	case QAT_SYMBOL:
		b.x.symbol=a.x.symbol;
		break;
	case QAT_INT:
		b.x.i=a.x.i;
		break;
	case QAT_FLOAT:
		b.x.f=a.x.f;
		break;
	case QAT_FUNC:
		b.x.func=a.x.func;
		break;
	case QAT_MENU:
		b.x.ptr=a.x.ptr;
		break;
	case QAT_STRING:
		b.x.string=Q_strdup(a.x.string);
		ERR_MEM(b.x.string);
		break;
	default:
		ERROR("QA_copy: unknown atom type %i",a.type);
	}
	return b;
}

qatom_t QA_List(void)
{
	qatom_t l;
	memset(&l,0,sizeof(qatom_t));
	return l;
}

void QA_List_Add(qatom_t *l,qatom_t a)
{
	if (!QAT_ISLIST(l->type))
		ERROR("QA_list_add: argument is not a list");
	l->x.l=Q_realloc(l->x.l,(l->type+1)*sizeof(qatom_t));
	ERR_MEM(l->x.l);
	l->x.l[l->type++]=a;
	ASSERT(QAT_ISLIST(l->type));
}


void QA_Describe(qatom_t *a,int ind)
{
	int i;
	if (QAT_ISLIST(a->type))
	{
		printf("%*c(\n",ind,' ');
		for (i=0;i<a->type;i++)
			QA_Describe(&a->x.l[i],ind+2);
		printf("%*c)\n",ind,' ');
		return;
	}
	for (i=0;i<ind;i++)
		fputc(' ',stdout);
	switch (a->type)
	{
	case QAT_SYMBOL:
//		printf("%*csymbol %i %s\n",ind,' ',a->x.symbol,QL_SymbolName(a->x.symbol));
		printf("%s\n",QL_SymbolName(a->x.symbol));
		break;
	case QAT_STRING:
//		printf("%*cstring %s\n",ind,' ',a->x.string);
		printf("\"%s\"\n",a->x.string);
		break;
	case QAT_INT:
		printf("%i\n",a->x.i);
		break;
	case QAT_FLOAT:
		printf("%g\n",a->x.f);
		break;
	case QAT_MENU:
		printf("menu %p\n",a->x.ptr);
	default:
		printf("!! unknown type %i !!\n",a->type);
		break;
	}
}

void QA_Print(qatom_t *a,FILE *f)
{
	if (QAT_ISLIST(a->type))
	{
		int i;
		fprintf(f,"(");
		for (i=0;i<a->type;i++)
		{
			QA_Print(&a->x.l[i],f);
			if (i<a->type-1) fprintf(f," ");
		}
		fprintf(f,")");
		return;
	}
	switch (a->type)
	{
	case QAT_SYMBOL:
		fprintf(f,"%s",QL_SymbolName(a->x.symbol));
		break;
	case QAT_INT:
		fprintf(f,"%i",a->x.i);
		break;
	case QAT_FLOAT:
		fprintf(f,"%#g",a->x.f);
		break;
	case QAT_STRING:
		fprintf(f,"\"%s\"",a->x.string);
		break;
	default:
		ERROR("Unknown atom type %i\n",a->type);
		break;
	}
}

int QA_Equal(qatom_t *a,qatom_t *b)
{
	if (a->type!=b->type) return 0; /* TODO: check int/float? */
	if (QAT_ISLIST(a->type))
	{
		int i;
		for (i=0;i<a->type;i++)
			if (!QA_Equal(&a->x.l[i],&b->x.l[i]))
				return 0;
		return 1;
	}
	switch (a->type)
	{
	case QAT_SYMBOL:
		return a->x.symbol==b->x.symbol;
	case QAT_INT:
		return a->x.i==b->x.i;
	case QAT_FLOAT:
/* TODO: epsilon? not really that important since this function is only used
when saving cvar:s */
		return a->x.f==b->x.f;
	case QAT_STRING:
		return !strcmp(a->x.string,b->x.string);
	
	default:
		ERROR("Unknown atom type %i\n",a->type);
		break;
	}
}

