#ifndef QATOM_H
#define QATOM_H

typedef struct qatom_s
{
	int type;

/* If type is <QA_BASE, it's a list and type is the number of elements in
the list. */
#define QAT_BASE      0x10000
#define QAT_ISLIST(t) ((t)<QAT_BASE)
#define QAT_UNDEF     (0+QAT_BASE)
#define QAT_SYMBOL    (1+QAT_BASE)
#define QAT_STRING    (2+QAT_BASE)
#define QAT_INT       (3+QAT_BASE)
#define QAT_FLOAT     (4+QAT_BASE)
#define QAT_FUNC      (5+QAT_BASE)
#define QAT_MENU      (6+QAT_BASE)
	union
	{
		struct qatom_s *l;
		int symbol;
		char *string;
		int i;
		float f;
		struct func_s *func;
		void *ptr;
	} x;
} qatom_t;


//#define QATOM_FAST

#ifdef QATOM_FAST

#define QLIST(lst,idx) (((lst)->x.l)+(idx))
#define QINT(at) ((at)->x.i)
#define QSTRING(at) ((at)->x.string)
#define QSYMBOL(at) ((at)->x.symbol)

#else

#define QLIST(lst,idx) \
	(!QAT_ISLIST((lst)->type)? \
		(Error_File(__FILE__,__FUNCTION__,__LINE__,"QLIST type error"),(qatom_t *)NULL) \
	:(((lst)->x.l)+(idx)))

#define QINT(at) \
	(((at)->type!=QAT_INT)? \
		(Error_File(__FILE__,__FUNCTION__,__LINE__,"QINT type error"),0) \
	:((at)->x.i))

#define QSTRING(at) \
	(((at)->type!=QAT_STRING)? \
		(Error_File(__FILE__,__FUNCTION__,__LINE__,"QSTRING type error"),NULL) \
	:((at)->x.string))

#define QSYMBOL(at) \
	(((at)->type!=QAT_SYMBOL)? \
		(Error_File(__FILE__,__FUNCTION__,__LINE__,"QSYMBOL type error"),0) \
	:((at)->x.symbol))

#endif


extern qatom_t qa_undef,qa_nil,qa_T;


void QA_Free(qatom_t a);
qatom_t QA_Copy(qatom_t a);

qatom_t QA_List(void);
void QA_List_Add(qatom_t *l,qatom_t a);

void QA_Describe(qatom_t *a,int ind);
void QA_Print(qatom_t *a,FILE *f);

int QA_Equal(qatom_t *a,qatom_t *b);

#endif

