#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "defines.h"
#include "types.h"

#include "ql.h"

#include "error.h"
#include "memory.h"
#include "qatom.h"
#include "qfunc.h"
#include "qsymbol.h"
#include "token.h"


#define A(x, y...) do { ERROR("Q_Load failure at line %i: " x,token_linenum , ##y); } while (0)

static qatom_t QL_LoadList(int gf);

static qatom_t QL_LoadAtom(void)
{
	qatom_t a;
//	printf("QL_LoadAtom()\n");
	if (!strcmp(token,"("))
	{
		return QL_LoadList(0);
	}
	if (token[0]=='"')
	{
		token[strlen(token)-1]=0;
		a.type=QAT_STRING;
		a.x.string=Q_strdup(&token[1]);
//		printf("string '%s'\n",a.x.string);
		return a;
	}
	if (token[0]>='0' && token[0]<='9')
	{
		int i;
		char *c;
		i=strtol(token,&c,0);
		if (!*c)
		{
			a.type=QAT_INT;
			a.x.i=i;
			return a;
		}
	}
	a.type=QAT_SYMBOL;
	a.x.symbol=QL_SymbolDef(token);
//	printf("symbol %i '%s'\n",a.x.symbol,token);
	return a;
}

static qatom_t QL_LoadList(int gf)
{
	qatom_t a;
//	printf("QL_LoadList(%i)\n",gf);
	if (gf) if (!TokenGet(1,-1)) A("Unexpected EOF");
//	printf("got '%s'\n",token);
	if (strcmp(token,"(")) A("Expected '('");
	a=QA_List();
	while (1)
	{
		if (!TokenGet(1,-1)) A("Unexpected EOF");
//		printf("loop got '%s'\n",token);
		if (!strcmp(token,")")) break;
//		printf("new elem\n");
		QA_List_Add(&a,QL_LoadAtom());
	}
//	printf("list done\n");
	return a;
}


#include "ql_config.h"

config_t **configs;
int num_configs;

config_t *config_quest,*config_fe;
config_t *config_fe_ui;


static config_t *FindConfig(const char *name)
{
	int i;
	for (i=0;i<num_configs;i++)
		if (!strcmp(configs[i]->name,name))
			return configs[i];
	return NULL;
}

static config_t *AddConfig(const char *name)
{
	config_t *c;
	c=FindConfig(name);
	if (c) return c;
	c=Q_malloc(sizeof(config_t));
	ABORT_MEM(c);
	memset(c,0,sizeof(config_t));
	c->name=Q_strdup(name);
	ABORT_MEM(c->name);
	return c;
}

void QL_Load(const char *filename)
{
	qatom_t l;
	int S_config,S_config_saved;

	S_config=QL_SymbolDef("config");
	S_config_saved=QL_SymbolDef("config_saved");

	if (!TokenFile(filename,TOKEN_C|TOKEN_MISC|TOKEN_STRING|TOKEN_NUMBER|TOKEN_LISPNAME,NULL))
	{
		ERROR("Can't open '%s'!",filename);
		return;
	}
	while (TokenAvailable(1))
	{
		l=QL_LoadList(1);
//		PrintList(l,1);
		if (l.type<1) A("Empty top-level list");
		if (l.x.l[0].type!=QAT_SYMBOL) A("First atom of top-level list not a symbol");
		if (l.x.l[0].x.symbol==S_config)
		{
			config_t *c;
			if (l.type<3) A("config: must have at least 3 arguments");
			c=AddConfig(QL_SymbolName(l.x.l[1].x.symbol));
			c->attributes=l.x.l[2];
			c->init.type=l.type-3;
			c->init.x=l.x;
			memmove(l.x.l,&l.x.l[3],sizeof(qatom_t)*(l.type-3));
			configs=Q_realloc(configs,sizeof(config_t *)*(num_configs+1));
			ERR_MEM(configs);
			configs[num_configs++]=c;
		}
		else
		if (l.x.l[0].x.symbol==S_config_saved)
		{
			int i;
			config_t *c;
			cvar_save_t *cv;
			if (l.type<2) A("config_saved: must have at least 2 arguments");
			c=AddConfig(QL_SymbolName(l.x.l[1].x.symbol));
			c->did_load=1;
			for (i=2;i<l.type;i++)
			{
				if (l.x.l[i].type!=2 || l.x.l[i].x.l[0].type!=QAT_SYMBOL)
					A("config_saved: invalid argument %i",i);
				c->cvs=Q_realloc(c->cvs,sizeof(cvar_save_t)*(c->num_cvs+1));
				ERR_MEM(c->cvs);
				cv=&c->cvs[c->num_cvs++];
				cv->symbol=l.x.l[i].x.l[0].x.symbol;
				cv->value=QA_Copy(l.x.l[i].x.l[1]);
			}
		}
		else
			A("Unknown top-level list function '%s'",QL_SymbolName(l.x.l[0].x.symbol));
	}
/*	PrintSTree(sizeof(int),1);
	printf("max_depth=%i num_items=%i item_depth=%g\n",max_depth,num_items,item_depth/(double)num_items);*/
	QL_SymbolBalance();
/*	max_depth=num_items=item_depth=0;
	PrintSTree(*(int *)&ql_symbols[0],1);
	printf("max_depth=%i num_items=%i item_depth=%g\n",max_depth,num_items,item_depth/(double)num_items);
	printf("have %i configs\n",num_configs);*/
/*	{
		int i;
		config_t *c;
		for (i=0;i<num_configs;i++)
		{
			c=configs[i];
			printf("config '%s'\n",c->name);
			printf("  attributes:\n");
			PrintList(&c->attributes,4);
			printf("  code:\n");
			PrintList(&c->init,4);
		}
	}*/
	TokenDone();
//	fflush(stdout);
//	QASSERT(0);
}


qatom_t QL_Eval(qatom_t a)
{
	qatom_t ret;
//	printf("** eval **\n");
//	PrintAtom(&a,1);
	if (QAT_ISLIST(a.type))
	{
		int i;
		func_t *f;
		qatom_t fa,al;

		if (a.type<1) ERROR("eval: empty list");
		fa=QL_Eval(a.x.l[0]);
		if (fa.type!=QAT_FUNC)
		{
			if (a.x.l[0].type==QAT_SYMBOL)
				ERROR("eval: no function '%s'",QL_SymbolName(a.x.l[0].x.symbol));
			else
				ERROR("eval: first element of list is not a function");
		}
		f=fa.x.func;
		al=QA_List();
		if (f->flags&FUNC_NOT_EVAL_ARGS)
		{
			for (i=1;i<a.type;i++)
				QA_List_Add(&al,QA_Copy(a.x.l[i]));
		}
		else
		{
			for (i=1;i<a.type;i++)
				QA_List_Add(&al,QL_Eval(a.x.l[i]));
		}

		ret=Func_Call(f,al);

		QA_Free(al);
	}
	else
	switch (a.type)
	{
	case QAT_SYMBOL:
		{
			ql_symbol_t *s=QL_Symbol(a.x.symbol);
			if (s && s->a)
				ret=QA_Copy(*s->a);
			else
				ret=a; /* TODO: should we really evaluate symbols to themselves? */
		}
		break;
	case QAT_STRING:
		ret=QA_Copy(a);
		break;
	case QAT_INT:
	case QAT_FLOAT:
		ret=a;
		break;
	default:
		ERROR("eval: unknown atom type %i",a.type);
	}
//	printf("evaluated to:\n");
//	PrintAtom(&ret,1);
	return ret;
}


static void QL_Run(qatom_t *l)
{
	int i;
	ASSERT(QAT_ISLIST(l->type));
	for (i=0;i<l->type;i++)
	{
		QA_Free(QL_Eval(l->x.l[i]));
	}
}


#include "plugin.h"

static config_t *loadconfig_current;

static void qf_plugin_load(qatom_t l)
{
	int i;
	plugin_t *p;
	if (!loadconfig_current)
		ERROR("plugin.load: can't load plugins when not in a configuration!");
	for (i=0;i<l.type;i++)
	{
		if (l.x.l[i].type!=QAT_SYMBOL) ERROR("plugin.load: arg %i not a symbol, type %i",i,l.x.l[i].type);
		p=Plugin_Lookup(QL_SymbolName(l.x.l[i].x.symbol));
		if (!p) ERROR("plugin.load: can't find plugin '%s'",QL_SymbolName(l.x.l[i].x.symbol));
		Plugin_Activate(p,loadconfig_current);
	}
}
static func_t F_plugin_load=
{"plugin.load",FUNC_VOID,{fva:qf_plugin_load},NULL,"s*"};


static void qf_set(qatom_t l)
{
	QL_SymbolSet(l.x.l[0].x.symbol,QL_Eval(l.x.l[1]));
}
static func_t F_set=
{"set",FUNC_VOID,{fva:qf_set},NULL,"s.",
FUNC_NOT_EVAL_ARGS
};

static void qf_quote(qatom_t *ret,qatom_t l)
{
	*ret=QA_Copy(l.x.l[0]);
}
static func_t F_quote=
{"quote",FUNC_ATOM,{faa:qf_quote},NULL,".",
FUNC_NOT_EVAL_ARGS
};

static void qf_config_load(qatom_t l)
{
	int i;
	for (i=0;i<l.type;i++)
	{
		if (QAT_ISLIST(l.x.l[i].type))
			qf_config_load(l.x.l[i]);
		else
		if (l.x.l[i].type==QAT_SYMBOL)
			QL_LoadConfig(QL_SymbolName(l.x.l[i].x.symbol));
		else
			ERROR("config.load: invalid argument type %i",l.x.l[i].type);
	}
}
static func_t F_config_load=
{"config.load",FUNC_VOID,{fva:qf_config_load},NULL,".*"};

static void qf_lambda(qatom_t *ret,qatom_t arg)
{ /* TODO: do this properly */
	func_t *f;
	qatom_t *c;
	ret->type=QAT_FUNC;
	f=ret->x.func=Q_malloc(sizeof(func_t));
	ERR_MEM(f);
	memset(f,0,sizeof(func_t));
	f->type=FUNC_QL;
	c=Q_malloc(sizeof(qatom_t));
	ERR_MEM(c);
	*c=QA_Copy(arg);
	f->data=c;
}
static func_t F_lambda=
{"lambda",FUNC_ATOM,{faa:qf_lambda},NULL,".+",
FUNC_NOT_EVAL_ARGS|FUNC_NOT_USER};


static void qf_define(qatom_t l)
{
	if (!loadconfig_current)
		ERROR("define: Can only define variables inside configurations");
	if (l.type==3)
		QL_SymbolCVar_Define(loadconfig_current,QSYMBOL(QLIST(&l,0)),QSTRING(QLIST(&l,2)));
	QL_SymbolCVar_Default(loadconfig_current,QSYMBOL(QLIST(&l,0)),*QLIST(&l,1));
}
static func_t F_define=
{"define",FUNC_VOID,{fva:qf_define},NULL,"s.c?",
FUNC_NOT_EVAL_ARGS|FUNC_NOT_USER};


static void qf_default(qatom_t l)
{
	if (!loadconfig_current)
		ERROR("default: Can only set default values inside configurations");
	QL_SymbolCVar_Default(loadconfig_current,QSYMBOL(QLIST(&l,0)),*QLIST(&l,1));
}
static func_t F_default=
{"default",FUNC_VOID,{fva:qf_default},NULL,"s.",
FUNC_NOT_EVAL_ARGS|FUNC_NOT_USER};



#include "err_catch.h"

static int S_configurations;

void QL_LoadConfig(const char *name)
{
	config_t *c;
	config_t *old;
	old=loadconfig_current;
	printf("QL_LoadConfig('%s')\n",name);
	c=FindConfig(name);
	if (!c) /* TODO: should be an 'important' warning */
	{
		Warning("No configuration '%s'!",name);
		return;
	}
	ERR_TRY
	{
		int i;
		cvar_save_t *cv;

		loadconfig_current=c;
		for (i=0,cv=c->cvs;i<c->num_cvs;i++,cv++)
			QL_SymbolCVar_Set(c,cv->symbol,cv->value);
		QL_Run(&c->init);
	}
	ERR_CATCH
	{
		loadconfig_current=old;
		ERROR("Error while loading configuration '%s'",c->name);
	}
	ERR_END
	c->active=1;
	c->did_load=1;
	loadconfig_current=old;
}


static void QL_UnloadConfig(config_t *cfg)
{
/* TODO:
Should:
- Unload all plugins loaded by this config (which may cause other things through hooks) (in order?)
- Save all cvar:s from this configuration
- Call a hk_config_unload hook so we can:
  - Remove menu items/menus/key bindings
  - More?
*/
	int sn=sizeof(int); /* The first symbol */
	ql_symbol_t *s;
	cvar_save_t *cv;

	ASSERT(cfg->active);

	Plugin_Deactivate(cfg);
	/* TODO: don't we need to free old cfg->cvs? */
	cfg->num_cvs=0;
	cfg->cvs=NULL;
	for (;sn<ql_symbols_size;sn=s->next)
	{
		s=QL_Symbol(sn);
		if (!s->cv) continue;
		if (s->cv->cfg!=cfg) continue;
		if (s->a &&
		    (s->cv->state==1 || s->cv->state==2 ||
		     s->cv->state==4 || s->cv->state==6 ||
		     !QA_Equal(&s->cv->vdefault,s->a)))
		{
			cfg->cvs=Q_realloc(cfg->cvs,sizeof(cvar_save_t)*(cfg->num_cvs+1));
			ABORT_MEM(cfg->cvs);
			cv=&cfg->cvs[cfg->num_cvs++];
			cv->symbol=sn;
			cv->value=QA_Copy(*s->a);
		}
		QL_SymbolCVar_Undef(sn);
	}

	cfg->active=0;
}


static func_t *funcs[]=
{&F_quote,&F_set,&F_plugin_load,&F_config_load,&F_lambda,
&F_define,&F_default
};

void QL_Init(void)
{
	Func_Register(NULL,funcs,sizeof(funcs));
	QL_SymbolDef("nil");
	qa_T.x.symbol=QL_SymbolDef("T");
	S_configurations=QL_SymbolDef("configurations");
	/* TODO: define constants? */
}

void QL_DidLoad(const char *fe_name)
{
	char buf[128];
	config_quest=FindConfig("Quest");
	if (!config_quest)
		ERROR("Can't find 'Quest' configuration!");

	snprintf(buf,sizeof(buf),"%s.settings",fe_name);
	config_fe=FindConfig(buf);
	if (!config_fe)
		ERROR("Can't find frontend configuration '%s'!",buf);
	QL_LoadConfig(buf);

	snprintf(buf,sizeof(buf),"%s.ui",fe_name);
	config_fe_ui=FindConfig(buf);
	if (!config_fe_ui)
		ERROR("Can't find frontend configuration '%s'!",buf);
}


static void QL_Save(const char *name) /* TODO */
{
	FILE *f;
	int i,j;
	config_t *c;
	f=fopen(name,"wt");
	if (!f) ERROR("Unable to save configuration in '%s' (%s)!",name,strerror(errno));

	fprintf(f,
		"/* Saved configuration values for Quest\n"
		"This file is automatically generated */\n\n");


	for (i=0;i<num_configs;i++)
	{
		if (!configs[i]->did_load)
			continue;
		ASSERT(!configs[i]->active);
		c=configs[i];
		fprintf(f,"(config_saved %s\n",c->name);
		for (j=0;j<c->num_cvs;j++)
		{
			fprintf(f," (%s ",QL_SymbolName(c->cvs[j].symbol));
			QA_Print(&c->cvs[j].value,f);
			fprintf(f,")\n");
		}
		fprintf(f,")\n\n");
	}

	fclose(f);
}


void QL_UnloadAll(void)
{
	int i;
	ql_symbol_t *s=QL_Symbol(S_configurations);
	config_t *c;
	ASSERT(QAT_ISLIST(s->a->type));
	for (i=s->a->type-1;i>=0;i--)
	{
		if (s->a->x.l[i].type!=QAT_SYMBOL)
			ERROR("Element %i of '%s' isn't a symbol!",
				i,s->name);
		c=FindConfig(QL_SymbolName(s->a->x.l[i].x.symbol));
		if (!c)
		{
			WARNING("Can't find configuration '%s'!",QL_SymbolName(s->a->x.l[i].x.symbol));
			continue;
		}
		if (!c->active)
		{
			WARNING("Configuration '%s' isn't loaded!",c->name);
			continue;
		}
		printf("unload '%s'\n",c->name);
		QL_UnloadConfig(c);
	}
	QL_UnloadConfig(config_fe_ui);
	QL_UnloadConfig(config_quest);
}

void QL_SaveConfig(void)
{ /* TODO: will this work? */
	QL_UnloadConfig(config_fe);
	QL_Save("cfgsave.ql");
}

