#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "qsymbol.h"

#include "error.h"
#include "memory.h"
#include "qfunc.h"
#include "ql.h"


static char *ql_symbols;
int ql_symbols_size;


#define QLS_RND(x) (((x)+(sizeof(int)-1))&~(sizeof(int)-1))
#define QLS_SIZE(x) (sizeof(ql_symbol_t)+strlen(x)+1)

int QL_SymbolDef(const char *name)
{
	int i,oi,j;
	ql_symbol_t *s;

	if (!strcmp(name,"nil")) return 0;

	if (!ql_symbols)
	{
		ql_symbols=Q_malloc(sizeof(int)+QLS_RND(QLS_SIZE(name)));
		ERR_MEM(ql_symbols);
		memset(ql_symbols,0,sizeof(int)+QLS_RND(QLS_SIZE(name)));
		*(int *)&ql_symbols[0]=sizeof(int);
		s=(ql_symbol_t *)&ql_symbols[sizeof(int)];
		strcpy(s->name,name);
		s->next=ql_symbols_size=sizeof(int)+QLS_RND(QLS_SIZE(name));
		return sizeof(int);
	}

	oi=j=0;
	for (i=*(int *)&ql_symbols[0];i;)
	{
		s=(ql_symbol_t *)&ql_symbols[i];
		if (!(j=strcmp(name,s->name)))
			return i;
		oi=i;
		if (j<0)
			i=s->lchild;
		else
			i=s->rchild;
	}
	ql_symbols=Q_realloc(ql_symbols,ql_symbols_size+QLS_RND(QLS_SIZE(name)));
	ERR_MEM(ql_symbols);
	i=ql_symbols_size;
	ql_symbols_size+=QLS_RND(QLS_SIZE(name));
	s=(ql_symbol_t *)&ql_symbols[oi];
	if (j<0)
		s->lchild=i;
	else
		s->rchild=i;
	s=(ql_symbol_t *)&ql_symbols[i];
	memset(s,0,sizeof(ql_symbol_t));
	strcpy(s->name,name);
	s->next=ql_symbols_size;
	return i;
}

ql_symbol_t *QL_Symbol(int s)
{
	if (!s) return NULL;
	return (ql_symbol_t *)&ql_symbols[s];
}

const char *QL_SymbolName(int s)
{
	if (!s) return "nil";
	return QL_Symbol(s)->name;
}


void QL_SymbolSetFunc(int sn,func_t *f)
{
	ql_symbol_t *s=QL_Symbol(sn);
	if (s->a)
		ERROR("'%s' redefined as function!",s->name);
	if (s->cv)
		ERROR("'%s' (cvar) redefined as function!",s->name);
	s->a=Q_malloc(sizeof(qatom_t));
	s->a->type=QAT_FUNC;
	s->a->x.func=f;
}

static void update_cvar(ql_symbol_t *s)
{
	config_var_t *c=s->cv;
	char buf[64];
	if (s->cv->s)
	{
		Q_free(s->cv->s);
		s->cv->s=NULL;
	}
	if (!s->a)
	{
		c->s=Q_strdup("");
		c->f=c->i=0;
		return;
	}
	switch (s->a->type)
	{
	case QAT_INT:
		c->f=c->i=s->a->x.i; /* TODO: set string? */
		snprintf(buf,sizeof(buf),"%i",c->i);
		c->s=Q_strdup(buf);
		ABORT_MEM(c->s);
		break;
	case QAT_FLOAT:
		c->i=c->f=s->a->x.f;
		if (c->i!=c->f)
			c->i=0;
		snprintf(buf,sizeof(buf),"%g",c->f);
		c->s=Q_strdup(buf);
		ABORT_MEM(c->s);
		break;
	case QAT_STRING:
		c->f=c->i=0;
		c->s=Q_strdup(s->a->x.string);
		ABORT_MEM(c->s);
		break;
	default:
		c->s=Q_strdup("");
		c->f=c->i=0;
		break;
	}
}

void QL_SymbolSet(int sn,qatom_t a)
{
	ql_symbol_t *s=QL_Symbol(sn);
	if (s->a)
	{
		if (s->a->type==QAT_FUNC)
			ERROR("function '%s' redefined",s->name);
		QA_Free(*s->a);
	}
	else
	{
		s->a=Q_malloc(sizeof(qatom_t));
		ERR_MEM(s->a);
	}
	*s->a=QA_Copy(a);
	if (s->cv)
		update_cvar(s);
}


void QL_SymbolCVar_Define(struct config_s *cfg,int symbol,const char *desc)
{
	ql_symbol_t *s=QL_Symbol(symbol);
	ASSERT(cfg);
	if (!s->cv)
	{
		s->cv=Q_malloc(sizeof(config_var_t));
		ERR_MEM(s->cv);
		memset(s->cv,0,sizeof(config_var_t));
		s->cv->cfg=cfg;
		s->cv->symbol=symbol;
		s->cv->description=desc;
		s->cv->state=1;
		update_cvar(s);
		return;
	}
	if (s->cv->cfg!=cfg)
		ERROR("cvar '%s' redefined",s->name);

	if (s->cv->state==0 || s->cv->state==1 || s->cv->state==4)
		ERROR("cvar '%s' in state %i",s->name,s->cv->state);

	s->cv->description=desc;
	if (s->cv->state==2)
		s->cv->state=4;
	else
		s->cv->state=0;
}

void QL_SymbolCVar_Default(struct config_s *cfg,int symbol,qatom_t vdefault)
{
	ql_symbol_t *s=QL_Symbol(symbol);
	ASSERT(cfg);
	if (!s->cv)
	{
		s->cv=Q_malloc(sizeof(config_var_t));
		ERR_MEM(s->cv);
		memset(s->cv,0,sizeof(config_var_t));
		s->cv->cfg=cfg;
		s->cv->symbol=symbol;
		s->cv->vdefault=QA_Copy(vdefault);
		s->cv->state=3;
		if (!s->a)
			QL_SymbolSet(symbol,vdefault);
		else
			update_cvar(s);
		return;
	}

	if (s->cv->cfg!=cfg && s->cv->cfg==config_quest && s->cv->state==0)
	{
		s->cv->state=5;
		s->cv->g_vdefault=s->cv->vdefault;
		s->cv->g_value=*s->a;
		Q_free(s->a);
		s->a=NULL;
		s->cv->vdefault=QA_Copy(vdefault);
		QL_SymbolSet(symbol,vdefault);
		return;
	}

	if (s->cv->cfg!=cfg || s->cv->state==0 || s->cv->state==3 || s->cv->state==5)
		ERROR("cvar '%s' in state %i",s->name,s->cv->state);

	s->cv->vdefault=QA_Copy(vdefault);

	if (s->cv->state==6)
		s->cv->state=5;
	else
	if (s->cv->state==1 && !s->a)
	{
		QL_SymbolSet(symbol,vdefault);
		s->cv->state=0;
	}
	else
	if (s->cv->state==2)
		s->cv->state=3;
	else
		s->cv->state=0;
}

void QL_SymbolCVar_Set(struct config_s *cfg,int symbol,qatom_t value)
{
	ql_symbol_t *s=QL_Symbol(symbol);
	ASSERT(cfg);
	if (!s->cv)
	{
		s->cv=Q_malloc(sizeof(config_var_t));
		ERR_MEM(s->cv);
		memset(s->cv,0,sizeof(config_var_t));
		s->cv->cfg=cfg;
		s->cv->symbol=symbol;
		s->cv->state=2;
		QL_SymbolSet(symbol,value);
		return;
	}
	if (cfg!=s->cv->cfg)
	{
		if (s->cv->cfg!=config_quest)
			ERROR("cvar '%s' in redefined",s->name);
		s->cv->cfg=cfg;
		s->cv->state=6;
		s->cv->g_vdefault=s->cv->vdefault;
		s->cv->g_value=*s->a;
		Q_free(s->a);
		s->a=NULL;
		QL_SymbolSet(symbol,value);
		return;
	}
	QL_SymbolSet(symbol,value);
	if (s->cv->state==1)
		s->cv->state=4;
}

void QL_SymbolCVar_Undef(int symbol)
{
	ql_symbol_t *s=QL_Symbol(symbol);
	ASSERT(s->cv);
	if (s->cv->state<=4)
	{
		if (s->cv->state==0 || s->cv->state==3)
			QA_Free(s->cv->vdefault);
		if (s->cv->s)
			Q_free(s->cv->s);
		Q_free(s->cv);
		s->cv=NULL;
		if (s->a)
		{
			QA_Free(*s->a);
			Q_free(s->a);
			s->a=NULL;
		}
	}
	else
	{
		QA_Free(*s->a);
		if (s->cv->state==5) QA_Free(s->cv->vdefault);
		s->cv->vdefault=s->cv->g_vdefault;
		s->cv->cfg=config_quest;
		s->cv->state=0;
		(*s->a)=s->cv->g_value;
		update_cvar(s);
	}
}


static void make_tree(int ofs,int *l,int nl)
{
	int i,o;
	if (!nl)
	{
		*(int *)&ql_symbols[ofs]=0;
		return;
	}
	i=l[o=nl/2];
	*(int *)&ql_symbols[ofs]=i;
	make_tree(i,l,o);
	make_tree(i+sizeof(int),&l[o+1],nl-o-1);
}

void QL_SymbolBalance(void)
{
	int *list;
	int nlist,slist;
#define MAX_STACK 64   /* TODO */
#define PUSH(x) stack[sdepth++]=(x)
#define POP() i=stack[--sdepth]
	int stack[MAX_STACK];
	int sdepth;
	int i;

	nlist=slist=0;
	list=NULL;

	sdepth=0;
	i=*(int *)&ql_symbols[0];
	for (;i || sdepth;)
	{
		if (i)
		{
			PUSH(i);
			i=*(int *)&ql_symbols[i];
		}
		else
		{
			POP();
			if (slist==nlist)
			{
				slist+=10;
				list=Q_realloc(list,sizeof(int)*slist);
				ERR_MEM(list);
			}
			list[nlist++]=i;
			i=*(int *)&ql_symbols[i+sizeof(int)];
		}
	}
	make_tree(0,list,nlist);
	Q_free(list);
}


#ifdef DEBUG_STUFF
/* debugging stuff */
static int max_depth,num_items,item_depth;
static void PrintSTree(int i,int ind)
{
	if (!i)
	{
		printf("%*cNULL\n",ind,' ');
		return;
	}
	if (ind>max_depth) max_depth=ind;
	num_items++;
	item_depth+=ind;
	printf("%*c%s %i\n",ind,' ',&ql_symbols[i+sizeof(int)*2],i);
	PrintSTree(*(int *)&ql_symbols[i],ind+1);
	PrintSTree(*(int *)&ql_symbols[i+sizeof(int)],ind+1);
}
#endif

