#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "defines.h"
#include "types.h"

#include "texdef.h"

#include "brush.h"
#include "config.h"
#include "game.h"
#include "status.h"
#include "undo.h"
#include "fe/message.h"
#include "fe/texture.h"


int SetTexture(texdef_t *t,const char *tname)
{
   strcpy(t->name,tname);
   return G_Texdef_Init(t);
}

int InitTexdef(texdef_t *t)
{
   return SetTexture(t,cv_texture_current->s);
}


void InitTexAlign(brush_normal_align_t *t)
{
   t->rotate=cv_texture_align_rotate->f;
   t->shift.x=cv_texture_align_shiftx->f;
   t->shift.y=cv_texture_align_shifty->f;
   t->scale.x=cv_texture_align_scalex->f;
   t->scale.y=cv_texture_align_scaley->f;
}

int InitFace(face_t *f)
{
   InitTexAlign(&f->a);
   return InitTexdef(&f->tex);
}


void SetTextureSel(sel_t *s,const char *tname)
{
   bsel_t *bs;
   brush_t *b;
   fsel_t *fs;
   
   int i;
   
   if (!tname)
   {
      S_NewMessage("Warning: Can't apply empty texture name!");
      return;
   }

   if (!s->bsel && !s->fsel)
   {
      S_NewMessage("SetTexture: Nothing selected!");
      return;
   }

   for (bs=s->bsel;bs;bs=bs->Next)
   {
      b=bs->Brush;

      U_BrushModify(b);
      
      if (b->bt->type==BRUSH_Q3_CURVE)
         SetTexture(&b->tex,tname);
      else
      if (b->bt->flags&BRUSH_F_NORMAL)
      {
         for (i=0;i<b->num_faces;i++)
            SetTexture(&b->faces[i]->tex,tname);
      }
   }

   for (fs=s->fsel;fs;fs=fs->Next)
   {
      U_BrushModify(fs->Brush);
      
      SetTexture(&fs->Brush->faces[fs->num]->tex,tname);
   }
}


static void td_move(brush_normal_align_t *a,int dir)
{
   switch (dir)
   {
   case MOVE_LEFT:  a->shift.x-=cv_edit_snap_size->i; break;
   case MOVE_RIGHT: a->shift.x+=cv_edit_snap_size->i; break;
   case MOVE_UP:    a->shift.y-=cv_edit_snap_size->i; break;
   case MOVE_DOWN:  a->shift.y+=cv_edit_snap_size->i; break;
   }
}

static void td_movecv(vec2_t *a,int dir,float mx,float my)
{
   switch (dir)
   {
   case MOVE_LEFT:  a->x-=mx; break;
   case MOVE_RIGHT: a->x+=mx; break;
   case MOVE_UP:    a->y-=my; break;
   case MOVE_DOWN:  a->y+=my; break;
   }
}

void Texdef_MoveSel(sel_t *s,int dir)
{
   bsel_t *bs;
   brush_t *b;
   fsel_t *fs;
   vsel_t *vs;
   int i;

   if (!(s->bsel || s->fsel))
   {
      S_NewMessage("Texdef_Move: Nothing selected!");
      return;
   }

   for (bs=s->bsel;bs;bs=bs->Next)
   {
      b=bs->Brush;
      if (b->bt->flags&BRUSH_F_NORMAL)
      {
         U_BrushModify(b);
         for (i=0;i<b->num_faces;i++)
            td_move(&b->faces[i]->a,dir);
      }
      else
      if (b->bt->type==BRUSH_Q3_CURVE)
      {
         texture_t *t;
         float mx,my;
         
         t=S_T_LoadSize(b->tex.name);
         if (t)
         {
            mx=cv_edit_snap_size->f/t->width;
            my=cv_edit_snap_size->f/t->height;
         }
         else
            mx=my=cv_edit_snap_size->f/64.0;
         
         U_BrushModify(b);
         if (!s->vsel)
         {
            for (i=0;i<b->num_verts;i++)
               td_movecv(&b->x.q3c.a[i],dir,mx,my);
         }
         else
         {
            for (vs=s->vsel;vs;vs=vs->Next)
            {
               if (vs->Brush!=b) continue;
               td_movecv(&b->x.q3c.a[vs->num],dir,mx,my);
            }
         }
      }
   }
   for (fs=s->fsel;fs;fs=fs->Next)
   {
      U_BrushModify(fs->Brush);
      td_move(&fs->Brush->faces[fs->num]->a,dir);
   }
}


static void td_scale(brush_normal_align_t *a,int dir)
{
   switch (dir)
   {
   case MOVE_LEFT:  a->scale.x-=0.1; break;
   case MOVE_RIGHT: a->scale.x+=0.1; break;
   case MOVE_UP:    a->scale.y-=0.1; break;
   case MOVE_DOWN:  a->scale.y+=0.1; break;
   }
}

void Texdef_ScaleSel(sel_t *s,int dir)
{
   bsel_t *bs;
   brush_t *b;
   fsel_t *fs;
   int i;

   for (bs=s->bsel;bs;bs=bs->Next)
   {
      b=bs->Brush;
      if (!(b->bt->flags&BRUSH_F_NORMAL))
         continue;
      U_BrushModify(b);
      for (i=0;i<b->num_faces;i++)
         td_scale(&b->faces[i]->a,dir);
   }
   for (fs=s->fsel;fs;fs=fs->Next)
   {
      U_BrushModify(fs->Brush);
      td_scale(&fs->Brush->faces[fs->num]->a,dir);
   }
}

